import collections
import operator
import os
from typing import Mapping

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import tqdm
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import MinMaxScaler


# The class is modified from https://github.com/nceglia/genevector
class GeneEmbedding(object):
    def __init__(self, embeddings: Mapping):
        self.embeddings = embeddings
        # self.context = dataset.data
        self.vector = []
        self.genes = []
        for gene in tqdm.tqdm(self.embeddings.keys()):
            self.vector.append(self.embeddings[gene])
            self.genes.append(gene)

    def read_embedding(self, filename):
        embedding = dict()
        lines = open(filename, "r").read().splitlines()[1:]
        for line in lines:
            vector = line.split()
            gene = vector.pop(0)
            embedding[gene] = np.array([float(x) for x in vector])
        return embedding

    def get_adata(self, resolution=20):
        mat = np.array(self.vector)
        np.savetxt(".tmp.txt", mat)
        gdata = sc.read_text(".tmp.txt")
        os.remove(".tmp.txt")
        gdata.obs.index = self.genes
        sc.pp.neighbors(gdata, use_rep="X")
        sc.tl.leiden(gdata, resolution=resolution)
        sc.tl.umap(gdata)
        return gdata

    def plot_similarities(self, gene, n_genes=10, save=None):
        df = self.compute_similarities(gene).head(n_genes)
        _, ax = plt.subplots(1, 1, figsize=(3, 6))
        sns.barplot(data=df, y="Gene", x="Similarity", palette="magma_r", ax=ax)
        ax.set_title("{} Similarity".format(gene))
        if save != None:
            plt.savefig(save)

    def plot_metagene(self, gdata, mg=None, title="Gene Embedding"):
        highlight = []
        labels = []
        clusters = collections.defaultdict(list)
        for x, y in zip(gdata.obs["leiden"], gdata.obs.index):
            clusters[x].append(y)
            if x == mg:
                highlight.append(str(x))
                labels.append(y)
            else:
                highlight.append("_Other")
        _labels = []
        for gene in labels:
            _labels.append(gene)
        gdata.obs["Metagene {}".format(mg)] = highlight
        _, ax = plt.subplots(1, 1, figsize=(8, 6))
        sc.pl.umap(gdata, alpha=0.5, show=False, size=100, ax=ax)
        sub = gdata[gdata.obs["Metagene {}".format(mg)] != "_Other"]
        sc.pl.umap(
            sub,
            color="Metagene {}".format(mg),
            title=title,
            size=200,
            show=False,
            add_outline=False,
            ax=ax,
        )
        for gene, pos in zip(gdata.obs.index, gdata.obsm["X_umap"].tolist()):
            if gene in _labels:
                ax.text(
                    pos[0] + 0.04,
                    pos[1],
                    str(gene),
                    fontsize=6,
                    alpha=0.9,
                    fontweight="bold",
                )
        plt.tight_layout()

    def plot_metagenes_scores(self, adata, metagenes, column, plot=None):
        plt.figure(figsize=(5, 13))
        matrix = []
        meta_genes = []
        cfnum = 1
        cfams = dict()
        for cluster, vector in metagenes.items():
            row = []
            cts = []
            for ct in set(adata.obs[column]):
                sub = adata[adata.obs[column] == ct]
                val = np.mean(sub.obs[str(cluster) + "_SCORE"].tolist())
                row.append(val)
                cts.append(ct)
            matrix.append(row)
            label = str(cluster) + "_SCORE: " + ", ".join(vector[:10])
            if len(set(vector)) > 10:
                label += "*"
            meta_genes.append(label)
            cfams[cluster] = label
            cfnum += 1
        matrix = np.array(matrix)
        df = pd.DataFrame(matrix, index=meta_genes, columns=cts)
        plt.figure()
        sns.clustermap(
            df,
            figsize=(5, 9),
            dendrogram_ratio=0.1,
            cmap="mako",
            yticklabels=True,
            standard_scale=0,
        )
        plt.tight_layout()
        if plot:
            plt.savefig(plot)

    def score_metagenes(self, adata, metagenes):
        for p, genes in metagenes.items():
            try:
                sc.tl.score_genes(adata, score_name=str(p) + "_SCORE", gene_list=genes)
                scores = np.array(adata.obs[str(p) + "_SCORE"].tolist()).reshape(-1, 1)
                scaler = MinMaxScaler()
                scores = scaler.fit_transform(scores)
                scores = list(scores.reshape(1, -1))[0]
                adata.obs[str(p) + "_SCORE"] = scores
            except Exception as e:
                adata.obs[str(p) + "_SCORE"] = 0.0

    def get_metagenes(self, gdata):
        metagenes = collections.defaultdict(list)
        for x, y in zip(gdata.obs["leiden"], gdata.obs.index):
            metagenes[x].append(y)
        return metagenes

    def compute_similarities(self, gene, subset=None, feature_type=None):
        if gene not in self.embeddings:
            return None
        # if feature_type:
        #     subset = []
        #     for gene in list(self.embeddings.keys()):
        #         if feature_type == self.context.feature_types[gene]:
        #             subset.append(gene)
        embedding = self.embeddings[gene]
        distances = dict()
        if subset:
            targets = set(list(self.embeddings.keys())).intersection(set(subset))
        else:
            targets = list(self.embeddings.keys())
        for target in targets:
            if target not in self.embeddings:
                continue
            v = self.embeddings[target]
            distance = float(
                cosine_similarity(
                    np.array(embedding).reshape(1, -1), np.array(v).reshape(1, -1)
                )[0]
            )
            distances[target] = distance
        sorted_distances = list(
            reversed(sorted(distances.items(), key=operator.itemgetter(1)))
        )
        genes = [x[0] for x in sorted_distances]
        distance = [x[1] for x in sorted_distances]
        df = pd.DataFrame.from_dict({"Gene": genes, "Similarity": distance})
        return df

    # def clusters(self, clusters):
    #     average_vector = dict()
    #     gene_to_cluster = collections.defaultdict(list)
    #     matrix = collections.defaultdict(list)
    #     total_average_vector = []
    #     for gene, cluster in zip(self.context.expressed_genes, clusters):
    #         if gene in self.embeddings:
    #             matrix[cluster].append(self.embeddings[gene])
    #             gene_to_cluster[cluster].append(gene)
    #             total_average_vector.append(self.embeddings[gene])
    #     self.total_average_vector = list(np.average(total_average_vector, axis=0))
    #     for cluster, vectors in matrix.items():
    #         xvec = list(np.average(vectors, axis=0))
    #         average_vector[cluster] = np.subtract(xvec, self.total_average_vector)
    #     return average_vector, gene_to_cluster

    def generate_weighted_vector(self, genes, markers, weights):
        vector = []
        for gene, vec in zip(self.genes, self.vector):
            if gene in genes and gene in weights:
                vector.append(weights[gene] * np.array(vec))
            if gene not in genes and gene in markers and gene in weights:
                vector.append(list(weights[gene] * np.negative(np.array(vec))))
        return list(np.sum(vector, axis=0))

    def generate_vector(self, genes):
        vector = []
        for gene, vec in zip(self.genes, self.vector):
            if gene in genes:
                vector.append(vec)
        assert len(vector) != 0, genes
        return list(np.average(vector, axis=0))

    def generate_weighted_vector(self, genes, weights):
        vector = []
        weight = []
        for gene, vec in zip(self.genes, self.vector):
            if gene in genes and gene in weights:
                vector.append(vec)
                weight.append(weights[gene])
        assert len(vector) != 0, genes
        return list(np.average(vector, axis=0, weights=weight))

    def cluster_definitions_as_df(self, top_n=20):
        similarities = self.cluster_definitions
        clusters = []
        symbols = []
        for key, genes in similarities.items():
            clusters.append(key)
            symbols.append(", ".join(genes[:top_n]))
        df = pd.DataFrame.from_dict({"Cluster Name": clusters, "Top Genes": symbols})
        return df

    @staticmethod
    def read_vector(vec):
        lines = open(vec, "r").read().splitlines()
        dims = lines.pop(0)
        vecs = dict()
        for line in lines:
            try:
                line = line.split()
                gene = line.pop(0)
                vecs[gene] = list(map(float, line))
            except Exception as e:
                continue
        return vecs, dims

    def get_similar_genes(self, vector):
        distances = dict()
        targets = list(self.embeddings.keys())
        for target in targets:
            if target not in self.embeddings:
                continue
            v = self.embeddings[target]
            distance = float(
                cosine_similarity(
                    np.array(vector).reshape(1, -1), np.array(v).reshape(1, -1)
                )[0]
            )
            distances[target] = distance
        sorted_distances = list(
            reversed(sorted(distances.items(), key=operator.itemgetter(1)))
        )
        genes = [x[0] for x in sorted_distances]
        distance = [x[1] for x in sorted_distances]
        df = pd.DataFrame.from_dict({"Gene": genes, "Similarity": distance})
        return df

    def generate_network(self, threshold=0.5):
        G = nx.Graph()
        a = pd.DataFrame.from_dict(self.embeddings).to_numpy()
        similarities = cosine_similarity(a.T)
        genes = list(self.embeddings.keys())
        similarities[similarities < threshold] = 0
        edges = []
        nz = list(zip(*similarities.nonzero()))
        for n in tqdm.tqdm(nz):
            edges.append((genes[n[0]], genes[n[1]]))
        G.add_nodes_from(genes)
        G.add_edges_from(edges)
        return G

    @staticmethod
    def average_vector_results(vec1, vec2, fname):
        output = open(fname, "w")
        vec1, dims = GeneEmbedding.read_vector(vec1)
        vec2, _ = GeneEmbedding.read_vector(vec2)
        genes = list(vec1.keys())
        output.write(dims + "\n")
        for gene in genes:
            v1 = vec1[gene]
            v2 = vec2[gene]
            meanv = []
            for x, y in zip(v1, v2):
                meanv.append(str((x + y) / 2))
            output.write("{} {}\n".format(gene, " ".join(meanv)))
        output.close()



from helical.models.base_models import HelicalRNAModel
from helical.models.helix_mrna.helix_mrna_config import HelixmRNAConfig
from datasets import Dataset
from transformers import BatchEncoding
from .modeling_helix_mrna import HelixmRNAPretrainedModel
from .helix_mrna_tokenizer import CharTokenizer
from .helix_mrna_pretrained_config import HelixmRNAPretrainedConfig
import torch
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm

import logging

logger = logging.getLogger(__name__)

class HelixmRNA(HelicalRNAModel):
    """Helix-mRNA Model.
    
    The Helix-mRNA Model is a transformer-based model that can be used to extract mRNA embeddings from mRNA sequences. 
    The model is based on the Mamba2 model, which is a transformer-based model trained on mRNA sequences. The model is available through this interface.
    
    Example
    -------
    ```python
    from helical import HelixmRNA, HelixmRNAConfig
    import torch

    device = "cuda" if torch.cuda.is_available() else "cpu"
    
    helix_mrna_config = HelimRNAConfig(batch_size=5, max_length=100, device=device)
    helix_mrna = HelixmRNA(configurer=helix_mrna_config)
    
    rna_sequences = ["EACUEGGG", "EACUEGGG", "EACUEGGG", "EACUEGGG", "EACUEGGG"]
    dataset = helix_mrna.process_data(rna_sequences)

    rna_embeddings = helix_mrna.get_embeddings(dataset)
    print("Helix_mRNA embeddings shape: ", rna_embeddings.shape)
    ```

    Parameters
    ----------
    configurer : HelixmRNAConfig
        The configuration object for the Helix-mRNA model.

    Notes
    ----------
    Helix_mRNA was trained using a character in between each codon of the mRNA sequence. 
    This is done to ensure that the model can learn the structure of the mRNA sequence. 
    Although it can take a standard RNA sequence as input, it is recommended to add the letter E between each codon of the mRNA sequence to get better embeddings.
    """
    default_configurer = HelixmRNAConfig()
    def __init__(self, configurer: HelixmRNAConfig = default_configurer):
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config
        
        self.model = HelixmRNAPretrainedModel.from_pretrained(self.config["model_name"])
        self.pretrained_config = HelixmRNAPretrainedConfig.from_pretrained(self.config["model_name"], trust_remote=True)
        self.tokenizer = CharTokenizer.from_pretrained(self.config["model_name"], trust_remote=True)
        logger.info("Helix-mRNA initialized successfully.")

    def process_data(self, sequences: str) -> Dataset:
        """Process the mRNA sequences and return a Dataset object.

        Parameters
        ----------
        sequences : str
            The mRNA sequences.

        Returns
        -------
        Dataset
            The dataset object.
        """
        self.ensure_rna_sequence_validity(sequences)
    
        tokenized_sequences = self.tokenizer(sequences, 
                                             return_tensors="pt", 
                                             padding="max_length", 
                                             truncation=True, 
                                             max_length=self.config['input_size'], 
                                             return_special_tokens_mask=True)
            
        dataset = Dataset.from_dict(tokenized_sequences)
        
        return dataset

    def get_embeddings(self, dataset: Dataset) -> np.ndarray:
        """Get the embeddings for the mRNA sequences.
        
        Parameters
        ----------
        dataset : HelixmRNADataset
            The dataset object.
        
        Returns
        -------
        np.ndarray
            The embeddings array.
        """
        dataloader = DataLoader(dataset, collate_fn=self._collate_fn, batch_size=self.config["batch_size"], shuffle=False)
        embeddings = []

        self.model.to(self.config["device"])

        progress_bar = tqdm(dataloader, desc="Getting embeddings")
        with torch.no_grad():
            for batch in progress_bar:
                input_ids = batch["input_ids"].to(self.config["device"])
                special_tokens_mask = batch["special_tokens_mask"].to(self.config["device"])
                attention_mask = 1-special_tokens_mask

                output = self.model(input_ids=input_ids, attention_mask=attention_mask)

                last_hidden_states = output[0]

                embeddings.append(last_hidden_states.cpu().numpy())

                del batch
                del output

        return np.concatenate(embeddings)
    

    def _collate_fn(self, batch):
        input_ids = torch.tensor([item["input_ids"] for item in batch])
        special_tokens_mask = torch.tensor([item["special_tokens_mask"] for item in batch])

        batch_dict = {
            "input_ids": input_ids,
            "special_tokens_mask": special_tokens_mask,
            "attention_mask": 1-special_tokens_mask
        }

        if "labels" in batch[0]:
            batch_dict["labels"] = torch.tensor([item["labels"] for item in batch])
            
        return BatchEncoding(batch_dict)
    


from helical.models.hyena_dna.model import HyenaDNAConfig
import pytest
import torch
from helical.models.hyena_dna.model import HyenaDNA
from helical.models.hyena_dna.hyena_dna_utils import HyenaDNADataset


@pytest.mark.skip(reason="Work in progress.")
@pytest.mark.parametrize("model_name, d_model, d_inner", [
    ("hyenadna-tiny-1k-seqlen", 128, 512),
    ("hyenadna-tiny-1k-seqlen-d256", 256, 1024)
])
def test_hyena_dna__valid_model_names(model_name, d_model, d_inner):
    """
    Test case for the HyenaDNA class initialization.

    Args:
        model_name (str): The name of the model.
        d_model (int): The dimensionality of the model.
        d_inner (int): The dimensionality of the inner layers.
    """
    configurer = HyenaDNAConfig(model_name=model_name)
    assert configurer.config["model_path"].name == f"{model_name}.ckpt"
    assert configurer.config["d_model"] == d_model
    assert configurer.config["d_inner"] == d_inner


@pytest.mark.skip(reason="Work in progress.")
@pytest.mark.parametrize("model_name", [
    ("wrong_name")
])
def test_hyena_dna__invalid_model_names(model_name):
    """
    Test case when an invalid model name is provided.
    Verifies that a ValueError is raised when an invalid model name is passed to the HyenaDNAConfig constructor.

    Parameters:
    - model_name (str): The invalid model name.

    Raises:
    - ValueError: If the model name is invalid.
    """
    with pytest.raises(ValueError):
        HyenaDNAConfig(model_name=model_name)


@pytest.mark.skip(reason="Work in progress.")
@pytest.mark.parametrize("input_sequence, expected_output", [
    # Valid DNA sequences
    ("", [0, 1]),
    ("A", [0, 7, 1]),
    ("CC", [0, 8, 8, 1]),
    # ("GGG", [0, 9, 9, 9, 1]),
    ("TTTT", [0, 10, 10, 10, 10, 1]),
    ("ACGTN", [0, 7, 8, 9, 10, 11, 1]),
    ("ACGT" * 256, [0] + [7, 8, 9, 10] * 256 + [1]),
    # Invalid sequences / sequences with uncertain 'N' nucleodites 
    ("BHIK", [0, 6, 6, 6, 6, 1]),
    ("ANNTBH", [0, 7, 11, 11, 10, 6, 6, 1]),

])
def test_hyena_dna_process_data(input_sequence, expected_output):
    """
    Test the process_data method of the HyenaDNA model.
    The input DNA sequence is tokenized and the output is compared to the expected output.

    Args:
        input_sequence (str): The input DNA sequence to be processed.
        expected_output (int): The expected output of the process_data method.

    Returns:
        None

    Raises:
        AssertionError: If the output of the process_data method does not match the expected output.
    """
    model = HyenaDNA()
    output = model.process_data([input_sequence])
    expected = torch.tensor([expected_output])
    assert torch.equal(output.sequences, expected)



from helical.models.base_models import HelicalRNAModel
from helical.models.mamba2_mrna.mamba2_mrna_config import Mamba2mRNAConfig
from .mamba2_mrna_tokenizer import CharTokenizer
from datasets import Dataset
from transformers import Mamba2Model, Mamba2Config, BatchEncoding
import torch
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm

import logging

logger = logging.getLogger(__name__)

class Mamba2mRNA(HelicalRNAModel):
    """Mamba2-mRNA Model.
    
    The Mamba2-mRNA Model is a transformer-based model that can be used to extract mRNA embeddings from mRNA sequences. 
    The model is based on the Mamba2 model, which is a transformer-based model trained on mRNA sequences. The model is available through this interface.
    
    Example
    ----------
    ```python
    from helical import Mamba2mRNA, Mamba2mRNAConfig
    import torch

    device = "cuda" if torch.cuda.is_available() else "cpu"

    input_sequences = ["ACUG"*20, "AUGC"*20, "AUGC"*20, "ACUG"*20, "AUUG"*20]

    mamba2_mrna_config = Mamba2mRNAConfig(batch_size=5, device=device)
    mamba2_mrna = Mamba2mRNA(configurer=mamba2_mrna_config)

    processed_input_data = mamba2_mrna.process_data(input_sequences)

    embeddings = mamba2_mrna.get_embeddings(processed_input_data)
    print(embeddings.shape)
    ```

    
    Parameters
    ----------
    configurer : Mamba2mRNAConfig
        The configuration object for the Mamba2-mRNA model.
    """
    default_configurer = Mamba2mRNAConfig()
    def __init__(self, configurer: Mamba2mRNAConfig = default_configurer):
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config
        
        self.model = Mamba2Model.from_pretrained(self.config["model_name"])
        self.pretrained_config = Mamba2Config.from_pretrained(self.config["model_name"], trust_remote=True)
        self.tokenizer = CharTokenizer.from_pretrained(self.config["model_name"], trust_remote=True)

        self.model.post_init()
        logger.info("Mamba2-mRNA initialized successfully.")

    def process_data(self, sequences: str) -> Dataset:
        """Process the mRNA sequences and return a Dataset object.

        Parameters
        ----------
        sequences : str
            The mRNA sequences.

        Returns
        -------
        Dataset
            The dataset object.
        """
        self.ensure_rna_sequence_validity(sequences)
    
        tokenized_sequences = self.tokenizer(sequences, 
                                             return_tensors="pt", 
                                             padding="max_length", 
                                             truncation=True, 
                                             max_length=self.config['input_size'], 
                                             return_special_tokens_mask=True)
        
        dataset = Dataset.from_dict(tokenized_sequences)
        
        return dataset

    def get_embeddings(self, dataset: Dataset) -> np.ndarray:
        """Get the embeddings for the mRNA sequences.
        
        Parameters
        ----------
        dataset : Dataset
            The dataset object.
        
        Returns
        -------
        np.ndarray
            The embeddings array.
        """
        dataloader = DataLoader(dataset, collate_fn=self._collate_fn, batch_size=self.config["batch_size"], shuffle=False)
        embeddings = []

        self.model.to(self.config["device"])

        progress_bar = tqdm(dataloader, desc="Getting embeddings")
        with torch.no_grad():
            for batch in progress_bar:
                input_ids = batch["input_ids"].to(self.config["device"])
                special_tokens_mask = batch["special_tokens_mask"].to(self.config["device"])
                attention_mask = batch["attention_mask"].to(self.config["device"])

                output = self.model(input_ids, special_tokens_mask=special_tokens_mask, attention_mask=attention_mask)

                last_hidden_states = output[0]

                if input_ids is not None:
                    batch_size, _ = input_ids.shape[:2]

                if self.pretrained_config.pad_token_id is None and batch_size > 1:
                    message = "Cannot handle batch sizes > 1 if no padding token is defined."
                    logger.error(message)
                    raise ValueError(message)

                if self.pretrained_config.pad_token_id is None:
                    sequence_lengths = -1
                else:
                    if input_ids is not None:
                        sequence_lengths = torch.eq(input_ids, self.pretrained_config.pad_token_id).int().argmax(-1) - 1
                        sequence_lengths = sequence_lengths % input_ids.shape[-1]
                        sequence_lengths = sequence_lengths.to(last_hidden_states.device)
                    else:
                        sequence_lengths = -1

                pooled_last_hidden_states = last_hidden_states[
                    torch.arange(batch_size, device=last_hidden_states.device), sequence_lengths
                ]
                
                # Take second last element from the output as last element is a special token
                embeddings.append(pooled_last_hidden_states.cpu().numpy())

                del batch
                del output

        return np.concatenate(embeddings)
    
    def _collate_fn(self, batch):
        input_ids = torch.tensor([item["input_ids"] for item in batch])
        special_tokens_mask = torch.tensor([item["special_tokens_mask"] for item in batch])

        batch_dict = {
            "input_ids": input_ids,
            "special_tokens_mask": special_tokens_mask,
            "attention_mask": 1-special_tokens_mask
        }

        if "labels" in batch[0]:
            batch_dict["labels"] = torch.tensor([item["labels"] for item in batch])
            
        return BatchEncoding(batch_dict)
    


import logging
from helical.models.hyena_dna.hyena_dna_config import HyenaDNAConfig
from helical.models.base_models import HelicalDNAModel
from tqdm import tqdm
from .hyena_dna_utils import HyenaDNADataset
from helical.models.hyena_dna.pretrained_model import HyenaDNAPreTrainedModel
import torch
from .standalone_hyenadna import CharacterTokenizer
from helical.utils.downloader import Downloader
from torch.utils.data import DataLoader
import numpy as np

LOGGER = logging.getLogger(__name__)

class HyenaDNA(HelicalDNAModel):
    """HyenaDNA model.
    This class represents the HyenaDNA model, which is a long-range genomic foundation model pretrained on context lengths of up to 1 million tokens at single nucleotide resolution.
    
    Example
    -------
    ```python
    from helical.models.hyena_dna.model import HyenaDNA, HyenaDNAConfig

    hyena_config = HyenaDNAConfig(model_name = "hyenadna-tiny-1k-seqlen-d256")
    model = HyenaDNA(configurer = hyena_config)   

    sequence = 'ACTG' * int(1024/4)

    tokenized_sequence = model.process_data(sequence)
    embeddings = model.get_embeddings(tokenized_sequence)

    print(embeddings.shape)
    ```

    Parameters
    ----------
    configurer : HyenaDNAConfig, optional, default=default_configurer
        The model configuration.

    Notes
    -----
    The link to the paper can be found [here](https://arxiv.org/abs/2306.15794. 
    We use the implementation from the [HyenaDNA](https://github.com/HazyResearch/hyena-dna) repository.
    """

    default_configurer = HyenaDNAConfig()

    def __init__(self, configurer: HyenaDNAConfig = default_configurer) -> None:    
        super().__init__()
        self.config = configurer.config

        downloader = Downloader()
        for file in self.config["list_of_files_to_download"]:
            downloader.download_via_name(file)

        self.model = HyenaDNAPreTrainedModel().from_pretrained(self.config)

        # create tokenizer
        self.tokenizer = CharacterTokenizer(
            characters=['A', 'C', 'G', 'T', 'N'],  # add DNA characters, N is uncertain
            model_max_length=self.config['max_length'] + 2,  # to account for special tokens, like EOS
            add_special_tokens=False,  # we handle special tokens elsewhere
            padding_side='left', # since HyenaDNA is causal, we pad on the left
        )

        # prep model and forward
        self.device = self.config['device']
        self.model.to(self.device)
        self.model.eval()
        LOGGER.info(f"Model finished initializing.")

    def process_data(self, sequence: list[str]) -> HyenaDNADataset:
        """Process the input DNA sequence.

        Parameters 
        ----------
        sequences : list[str]
            The input DNA sequences to be processed.

        Returns
        -------
        HyenaDNADataset
            Containing processed DNA sequences.

        """
        LOGGER.info(f"Processing data")
        processed_sequences = []
        for seq in tqdm(sequence, desc="Processing sequences"):
            tok_seq = self.tokenizer(seq)
            tok_seq_input_ids = tok_seq["input_ids"]

            tensor = torch.LongTensor(tok_seq_input_ids)
            tensor = tensor.to(self.device)
            processed_sequences.append(tensor)

        dataset = HyenaDNADataset(torch.stack(processed_sequences))
        LOGGER.info(f"Data processing finished.")

        return dataset

    def get_embeddings(self, dataset: HyenaDNADataset) -> torch.Tensor:
        """Get the embeddings for the tokenized sequence.

        Parameters
        ----------
        dataset : HyenaDNADataset
            The output dataset from `process_data`.

        Returns
        ----------
        np.ndarray
            The embeddings for the tokenized sequence in the form of a numpy array.

        """
        LOGGER.info(f"Inference started")

        train_data_loader = DataLoader(dataset, batch_size=self.config["batch_size"])
        with torch.inference_mode():
            embeddings = []
            for input_data in tqdm(train_data_loader, desc="Getting embeddings"):
                input_data = input_data.to(self.device)
                embeddings.append(self.model(input_data).detach().cpu().numpy())
        
        # output = torch.stack(embeddings)
        # other_dims = output.shape[2:]

        # reshaped_tensor = output.view(-1, *other_dims)
        return np.vstack(embeddings)



from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout, BatchNormalization
from tensorflow.keras.callbacks import TensorBoard
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.utils import to_categorical
from tensorflow_addons.metrics import F1Score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
import numpy as np
import datetime
from helical.models.hyena_dna.hyena_dna_config import HyenaDNAConfig    
import json

configurer = HyenaDNAConfig(model_name="hyenadna-tiny-1k-seqlen-d256")

from datasets import get_dataset_config_names
from datasets import load_dataset
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import f1_score
import sklearn

configs = get_dataset_config_names("InstaDeepAI/nucleotide_transformer_downstream_tasks")
result = []
for i, label in enumerate(configs):
    print("Processing:", label, i+1, "/", len(configs))
    dataset = load_dataset("InstaDeepAI/nucleotide_transformer_downstream_tasks", label)
    num_classes = len(set(dataset["train"]["label"]))


    x = np.load(f"data/train/x_{label}_norm_256.npy")
    y = np.load(f"data/train/y_{label}_norm_256.npy")
    # # One-hot encode the labels
    encoder = LabelEncoder()
    y_encoded = encoder.fit_transform(y)
    y_encoded = to_categorical(y_encoded, num_classes=num_classes)

    # X_train, X_test, y_train, y_test = train_test_split(x, y_encoded, test_size=0, random_state=42)

    input_shape = (configurer.config['d_model'],)

    # define the model
    head_model = Sequential()
    head_model.add(Dense(256, activation='relu', input_shape=input_shape))
    head_model.add(Dropout(0.4)) 
    head_model.add(Dense(128, activation='relu'))
    head_model.add(Dropout(0.4))
    head_model.add(Dense(128, activation='relu'))
    head_model.add(Dropout(0.4))  
    head_model.add(Dense(64, activation='relu'))
    head_model.add(Dropout(0.4))  
    head_model.add(Dense(num_classes, activation='softmax'))

    # compile the model
    optimizer = Adam(learning_rate=0.001)
    f1_score = F1Score(num_classes, average='macro')

    head_model.compile(loss='categorical_crossentropy', optimizer=optimizer, metrics=f1_score)

    # Setup callbacks
    log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    tensorboard_callback = TensorBoard(log_dir=log_dir, histogram_freq=1)
    history = head_model.fit(x, y_encoded, epochs=40, batch_size=64, callbacks=[tensorboard_callback])

    X_unseen = np.load(f"data/test/x_{label}_norm_256.npy")
    y_unseen = np.load(f"data/test/y_{label}_norm_256.npy")

    predictions_nn = head_model.predict(X_unseen)
    y_pred = np.argmax(predictions_nn, axis=1)

    # accuracy.append(sum(np.equal(y_pred, y_unseen))/len(y_unseen))
    # Matthews correlation coefficient (MCC) for enhancer and epigenetic datasets
    # F1-score for promoter and splice site datasets
    if label in ["H4ac", 
                 "H3K36me3", 
                 "H3", 
                 "H4", 
                 "H3K4me3", 
                 "H3K4me1", 
                 "H3K14ac", 
                 "enhancers_types", 
                 "H3K79me3", 
                 "H3K4me2", 
                 "enhancers", 
                 "H3K9ac"]:
        result.append(matthews_corrcoef(y_unseen, y_pred))
        
    else:
        if num_classes == 2:
            result.append(sklearn.metrics.f1_score(y_unseen, y_pred, average='binary'))
        else:
            # I'm not 100% sure what the average parameter should be here. I dont think the paper specifies it
            # splice_sites_all has 3 classes
            result.append(sklearn.metrics.f1_score(y_unseen, y_pred, average='weighted'))


with open("file.json", 'w') as f:
    json.dump(result, f, indent=2) 
    json.dump(configs, f, indent=2) 







import pytest
import numpy as np
from datasets import Dataset, Features, Value, Sequence
from helical.utils import get_anndata_from_hf_dataset

def create_mock_dataset(gene_names: str):
    data = {
        'raw_counts': [
            [1, 2, 3, 2],
            [98],
            [72, 19]
        ],
        'rows': [
            [0, 1, 2, 3],
            [0],
            [1, 3]
        ],
        'obs1': [10, 20, 30],
        'obs2': [40, 50, 60],
        'size': [4, 4, 4]
    }
    features = Features({
        'raw_counts': Sequence(Value('uint32'), -1, gene_names),
        'rows': Sequence(Value('uint32')),
        'obs1': Value('int64'),
        'obs2': Value('int64'),
        'size': Value('uint32')
    })
    dataset = Dataset.from_dict(data, features=features)
    return dataset

def test_get_anndata_from_hf_dataset():
    dataset = create_mock_dataset("gene1,gene2,gene3,gene4")
    ann_data = get_anndata_from_hf_dataset(dataset)
    
    assert ann_data.shape == (3, 4)

    # assert that observation names are correct (ie. no 'rows', 'raw_counts', or 'size')
    assert list(ann_data.obs.columns) == ['obs1', 'obs2']

    # assert that gene names are converted to uppercase 
    assert list(ann_data.var_names) == ['GENE1', 'GENE2', 'GENE3', 'GENE4']
    assert list(ann_data.var['gene_name']) == ['GENE1', 'GENE2', 'GENE3', 'GENE4']
    
    # assert that counts are placed in the correct positions
    assert np.array_equal(ann_data.X.toarray(), np.array([[1, 2, 3, 2], [98, 0, 0, 0], [0, 72, 0, 19]]))

def test_get_anndata_from_hf_dataset_mismatched_gene_names():
    dataset = create_mock_dataset("gene1,gene2")
    
    with pytest.raises(ValueError):
        get_anndata_from_hf_dataset(dataset)


from helical.benchmark.benchmark import evaluate_classification, evaluate_integration
from helical.models.geneformer.model import Geneformer
from helical.models.geneformer.geneformer_config import GeneformerConfig
from helical.models.scgpt.model import scGPT
from helical.models.scgpt.scgpt_config import scGPTConfig
from helical.models.uce.model import UCE
from helical.models.uce.uce_config import UCEConfig
from helical.models.classification.svm import SupportVectorMachine
from helical.models.classification.classifier import Classifier
from helical.utils import get_anndata_from_hf_dataset
from datasets import load_dataset
import anndata as ad
from omegaconf import DictConfig
import hydra
import json
import logging
import os
from anndata import AnnData
import scanpy as sc
import scanpy.external as sce

LOGGER = logging.getLogger(__name__)

class Scanorama():
    def __init__(self, batch_key: str):
        self.key = batch_key

    def process_data(self, data, **kwargs) -> AnnData:
        data = data[data.obs[self.key].sort_values().index]
        return data

    def get_embeddings(self, data: AnnData):
        sc.pp.recipe_zheng17(data)
        sc.pp.pca(data)
        sce.pp.scanorama_integrate(data, self.key, verbose=1)
        return data.obsm["X_scanorama"]


def write_to_json(evaluations: dict, path: str, file_name: str) -> None:
    """
    Write the evaluations to a JSON file.

    Parameters:
    -----------
        evaluations: dict
            The evaluations to be written to the JSON file.
        path: str
            The path where the JSON file will be saved.
        file_name: str
            The name of the JSON file.
    """
    # Serializing json
    json_object = json.dumps(evaluations, indent=4)
    
    os.makedirs(os.path.dirname(path), exist_ok=True)  

    # Writing to sample.json
    with open(f"{path}{file_name}.json", "w+") as outfile:
        outfile.write(json_object)

def get_model(model_name: str, data_cfg: DictConfig, device: str = "cpu"):
    if model_name == "geneformer":
        configurer = GeneformerConfig(device=device, batch_size=5)
        return Geneformer(configurer)
    elif model_name == "scgpt":
        configurer = scGPTConfig(device=device)
        return scGPT(configurer)
    elif model_name == "uce":
        configurer = UCEConfig(device=device)
        return UCE(configurer)
    elif model_name == "scanorama":
        return Scanorama(data_cfg["batch_key"])
    else:
        raise ValueError(f"Model {model_name} not found.")

def run_classification_example(data: AnnData, probing_models: list[str], data_cfg: DictConfig, head_cfg: DictConfig, device: str = "cpu"):
    """
    This only runs with probing models. Meaning that a Classifier head is trained on top of the embeddings generated by a foundation model.

    Parameters
    ----------
    data : AnnData
        The data to be classified.
    probing_models : list[str]
        The models to be used for classification.
    data_cfg : DictConfig
        The configuration of the data. It must contain the keys "gene_names" and "label_key".
    head_cfg : DictConfig
        The configuration what model head to use and how it is configured.
    device : str
        The device to run the models on. Default is "cpu".
    """

    train_data = data[:int(len(data)*0.8)]
    eval_data = data[int(len(data)*0.8):]
    evaluations_all = {}
    
    for model_name in probing_models:
        model = get_model(model_name, data_cfg, device)
        model_c = Classifier().train_classifier_head(train_data,
                                                        model, 
                                                        SupportVectorMachine(**head_cfg), 
                                                        gene_names=data_cfg["gene_names"],
                                                        labels_column_name=data_cfg["label_key"],
                                                        test_size=0.1,
                                                        random_state=42) 
        evaluations = evaluate_classification([model_c], eval_data, data_cfg["label_key"])
        evaluations_all.update({model_name: evaluations})
        # save outputs
        model_c.trained_task_model.save(f"{data_cfg['base_dir']}/{data_cfg['name']}/cell_type_annotation/{model_name}/")        
        write_to_json(evaluations, f"{data_cfg['base_dir']}/{data_cfg['name']}/cell_type_annotation/", f"classification_evaluations_{model_name}")

        # memory management
        del model_c, model

def run_integration_example(data: AnnData, models: list[str], data_cfg: DictConfig, integration_cfg: DictConfig, device: str = "cpu"):
    """
    Run integration evaluation for the specified models.

    Parameters
    ----------
    data : AnnData
        The data to be integrated.
    models : list[str]
        The models to be used for integration.
    data_cfg : DictConfig
        The configuration of the data. 
    integration_cfg : DictConfig
        The configuration of the integration.
    device : str
        The device to run the models on. Default is "cpu".
    """

    for model_name in models:
        model = get_model(model_name, data_cfg, device)
        evaluations = evaluate_integration([(model, model_name)], data, data_cfg, integration_cfg)
        write_to_json(evaluations, f"{data_cfg['base_dir']}/{data_cfg['name']}/integration/", f"integration_evaluations_{model_name}")
        del model

@hydra.main(version_base=None, config_path=".", config_name="config")
def benchmark(cfg: DictConfig) -> None:

    cfg["device"] = "cuda"
    dataset = "yolksac"

    data_cfg = cfg["data"][dataset]
    head_cfg = cfg["svm"]
    integration_cfg = cfg["integration"]

    # either load via huggingface
    # hf_dataset = load_dataset(data_cfg["path"], split="train[:10%]", trust_remote_code=True, download_mode="reuse_cache_if_exists")
    # data = get_anndata_from_hf_dataset(hf_dataset)[:10]

    # or load directly
    data = ad.read_h5ad("./yolksac_human.h5ad")[:10]
    data.obs[data_cfg["label_key"]] = data.obs[data_cfg["label_key"]].astype("category")

    # set gene names. for example if the index is the ensemble gene id 
    # data.var_names = data.var["feature_name"]

    run_classification_example(data, ["scgpt", "geneformer"], data_cfg, head_cfg, device=cfg["device"])
    # run_integration_example(data, ["scgpt", "geneformer", "scanorama"], data_cfg, integration_cfg, device=cfg["device"])
    LOGGER.info("Benchmarking done.")

if __name__ == "__main__":
    benchmark()


import os
import logging

logging.captureWarnings(True)

class InfoAndErrorFilter(logging.Filter):
    def filter(self, record):
        return record.levelno in (logging.INFO, logging.ERROR)

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logger = logging.getLogger()
logger.setLevel(logging.INFO)

handler = logging.StreamHandler()
handler.setLevel(logging.INFO) 

handler.addFilter(InfoAndErrorFilter())

formatter = logging.Formatter('%(levelname)s:%(name)s:%(message)s')
handler.setFormatter(formatter)

logger.addHandler(handler)

os.environ['TRANSFORMERS_VERBOSITY'] = 'error'


from .models.uce.model import UCEConfig, UCE
from .models.uce.fine_tuning_model import UCEFineTuningModel
from .models.geneformer.model import Geneformer,GeneformerConfig
from .models.geneformer.fine_tuning_model import GeneformerFineTuningModel
from .models.scgpt.model import scGPT, scGPTConfig
from .models.scgpt.fine_tuning_model import scGPTFineTuningModel
from .models.hyena_dna.model import HyenaDNA, HyenaDNAConfig
from .models.hyena_dna.fine_tuning_model import HyenaDNAFineTuningModel
from .models.helix_mrna import HelixmRNA, HelixmRNAConfig
from .models.helix_mrna import HelixmRNAFineTuningModel
from .models.mamba2_mrna import Mamba2mRNA, Mamba2mRNAConfig, Mamba2mRNAFineTuningModel




from helical.benchmark.benchmark import evaluate_classification, evaluate_integration
from helical.models.classification.classifier import Classifier
from helical.models.classification.neural_network import NeuralNetwork
from helical.models.scgpt.model import scGPT
import anndata as ad
import numpy as np
import pandas as pd
from anndata import AnnData

scgpt = scGPT()
data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
config = {
    "data": {
        "batch_key": "batch",
        "label_key": "str_labels",
        "gene_names": "index"
    },
    "integration": {
        "scib": {
            "isolated_labels_asw_": False,
            "silhouette_": True,
            "hvg_score_": False,
            "graph_conn_": True,
            "pcr_": True,
            "isolated_labels_f1_": False,
            "trajectory_": False,
            "nmi_": True,  # use the clustering bias to the best matching
            "ari_": True,  # use the clustering bias to the best matching
            "cell_cycle_": False,
            "kBET_": False,  # kBET return nan sometimes need to examine
            "ilisi_": False,
            "clisi_": False
        }
    }
}

generator = np.random.default_rng(seed=42)
data.obs["batch"] = generator.choice([0, 1], size=data.shape[0])
data.obs["str_labels"] = pd.Categorical(['type1', 'type2'] * (data.shape[0] // 2))
data.obsm["X_scgpt"] = np.zeros((data.shape[0], 10))

def assert_near_exact(x, y, diff=1e-5):
    assert abs(x - y) <= diff, f"{x} != {y} with error margin {diff}"

def test_evaluate_classification(mocker):
    head = NeuralNetwork()

    # Mocking the get_embeddings method by returning a zero matrix
    mocker.patch.object(scgpt, 'get_embeddings')
    scgpt.get_embeddings.return_value = np.zeros((data.shape[0], 10))
    
    # Mocking the training process by returning the untrained head
    mocker.patch.object(head, 'train')
    head.train.return_value = head

    # Mocking the prediction process by returning the untrained head
    mocker.patch.object(head, 'predict')
    head.predict.return_value = np.array(['type1', 'type2'] * (data.shape[0] // 2))

    scgpt_nn_c = Classifier().train_classifier_head(data, scgpt, head, gene_names = "gene_symbols", labels_column_name = "cell_type")
    evaluations = evaluate_classification([scgpt_nn_c, scgpt_nn_c], data, "cell_type")
    assert evaluations == {'scGPT with NeuralNetwork': {'Accuracy': 1.0, 'Precision': 1.0, 'F1': 1.0, 'Recall': 1.0}, 
                           'scGPT with NeuralNetwork': {'Accuracy': 1.0, 'Precision': 1.0, 'F1': 1.0, 'Recall': 1.0}}
    
def test_evaluate_integration(mocker):
    """
    Test that the integration evaluation function returns the expected results.
    We follow the approach from scGPT to evaluate the integration based on the bioloical conservation
    and the batch correction. The metrics we are interested in are:
    - ARI (Adjusted Rand Index)
    - NMI (Normalized Mutual Information)
    - ASW (Average Silhouette Width)
    - Graph Connectivity
    """

    # Mocking the get_embeddings method by returning a zero matrix
    mocker.patch.object(scgpt, 'get_embeddings')
    scgpt.get_embeddings.return_value = np.zeros((data.shape[0], 10))

    evaluations = evaluate_integration([(scgpt, "scgpt")], data, config["data"], config["integration"])
    
    # scgpt
    assert_near_exact(evaluations["scgpt"]["BATCH"]["ASW_batch"], 1.0)
    assert_near_exact(evaluations["scgpt"]["BATCH"]["Graph_Conn"], 1.0)
    
    assert_near_exact(evaluations["scgpt"]["BIO"]["ARI_cell"], 0.0)
    assert_near_exact(evaluations["scgpt"]["BIO"]["NMI_cell"], 0.2616480412956257)
    assert_near_exact(evaluations["scgpt"]["BIO"]["ASW_cell"], 0.5)

def test_integration_with_custom_model():
    class CustomWrapper():
        def process_data(self, data, **kwargs) -> AnnData:
            return data

        def get_embeddings(self, data: AnnData):
            return np.zeros((data.shape[0], 10))
        
        def train(self):
            pass # TODO

    custom_model = CustomWrapper()
    evaluations = evaluate_integration([(custom_model, "custom_model")], data, config["data"], config["integration"])
  
    # custom_model
    assert_near_exact(evaluations["custom_model"]["BATCH"]["ASW_batch"], 1.0)
    assert_near_exact(evaluations["custom_model"]["BATCH"]["Graph_Conn"], 1.0)

    assert_near_exact(evaluations["custom_model"]["BIO"]["ARI_cell"], 0.0)
    assert_near_exact(evaluations["custom_model"]["BIO"]["NMI_cell"], 0.2616480412956257)
    assert_near_exact(evaluations["custom_model"]["BIO"]["ASW_cell"], 0.5)



from typing import Literal, Optional
from helical.models.base_models import HelicalBaseFineTuningHead, HelicalBaseFineTuningModel
from helical.models.hyena_dna import HyenaDNA, HyenaDNAConfig
from torch import optim
import torch
from torch.nn.modules import loss
from .hyena_dna_utils import HyenaDNADataset
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import get_scheduler
import logging
import numpy as np

logger = logging.getLogger(__name__)

class HyenaDNAFineTuningModel(HelicalBaseFineTuningModel, HyenaDNA):
    """HyenaDNA fine-tuning model.

    This class represents the HyenaDNA fine-tuning model, which is a long-range genomic foundation model pretrained on context lengths of up to 1 million tokens at single nucleotide resolution.

    Example
    ----------
    ```python
    from datasets import load_dataset
    from helical import HyenaDNAConfig, HyenaDNAFineTuningModel
    import torch

    device = "cuda" if torch.cuda.is_available() else "cpu"

    # Load a Hugging Face dataset and task type
    ds = load_dataset("dataset", "task")

    # Define the desired configs
    config = HyenaDNAConfig(device=device, batch_size=10)

    # Define the fine-tuning model with the configs we instantiated above
    hyena_fine_tune = HyenaDNAFineTuningModel(config, "classification", number_unique_outputs)

    # Prepare the sequences for input to the model
    input_dataset = hyena_fine_tune.process_data(ds["train"]["sequence"])

    # train the fine-tuning model on some downstream task
    hyena_fine_tune.train(input_dataset, ds["train"]["label"])
    ```
    
    Parameters
    ----------
    hyena_config : HyenaDNAConfig
        The HyenaDNA configs for fine-tuning model, the same configs that would be used to instantiate the standard HyenaDNA model.
    fine_tuning_head : Literal["classification", "regression"]|HelicalBaseFineTuningHead
        The fine-tuning head that is appended to the model. This can either be a string (options available: "classification", "regression") specifying the task or a custom fine-tuning head inheriting from HelicalBaseFineTuningHead.
    output_size : Optional[int], default=None
        The output size of the fine-tuning head. This is required if a predefined head is selected.
    
    Methods
    -------
    train(train_input_data: HyenaDNADataset, train_labels: list[int], validation_input_data: HyenaDNADataset = None, validation_labels: list[int] = None, optimizer: torch.optim, optimizer_params: dict, loss_function: torch.nn.modules.loss, epochs: int, lr_scheduler_params: dict = None)
        Fine-tunes the Hyena-DNA model with different head modules.
    get_outputs(input_data: HyenaDNADataset) -> np.ndarray
        Get the outputs of the fine-tuned model.
    """
    def __init__(
            self, 
            hyena_config: HyenaDNAConfig, 
            fine_tuning_head: Literal["classification"]|HelicalBaseFineTuningHead, 
            output_size: Optional[int]=None):
        HelicalBaseFineTuningModel.__init__(self, fine_tuning_head, output_size)
        HyenaDNA.__init__(self, hyena_config)

        self.fine_tuning_head.set_dim_size(self.config["d_model"])

    def _forward(self, x):
        x = self.model(x)
        x = torch.mean(x, dim=1)
        x = self.fine_tuning_head(x)
        return x

    def train(        
        self,
        train_input_data: HyenaDNADataset,
        train_labels: list[int],     
        validation_input_data: HyenaDNADataset = None,
        validation_labels: list[int] = None,
        optimizer: optim = optim.AdamW,
        optimizer_params: dict = {'lr': 0.0001}, 
        loss_function: loss = loss.CrossEntropyLoss(), 
        epochs: int = 1,
        lr_scheduler_params: Optional[dict] = None):
        """Fine-tunes the Hyena-DNA model with different head modules. 

        Parameters
        ----------
        train_input_data : HyenaDNADataset
            A helical Hyena-DNA processed dataset for fine-tuning
        train_labels : list[int]
            The labels for the training data. These should be stored as unique per class integers.
        validation_input_data : HyenaDNADataset, default=None
            A helical Hyena-DNA processed dataset for per epoch validation. If this is not specified, no validation will be performed.
        validation_labels : list[int], default=None
            The labels for the validation data. These should be stored as unique per class integers.
        optimizer : torch.optim, default=torch.optim.AdamW
            The optimizer to be used for training.
        optimizer_params : dict
            The optimizer parameters to be used for the optimizer specified. This list should NOT include model parameters.
            e.g. optimizer_params = {'lr': 0.0001}
        loss_function : torch.nn.modules.loss, default=torch.nn.CrossEntropyLoss()
            The loss function to be used.
        epochs : int, optional, default=10
            The number of epochs to train the model for.
        lr_scheduler_params : dict, default=None
            The learning rate scheduler parameters for the transformers get_scheduler method. The optimizer will be taken from the optimizer input and should not be included in the learning scheduler parameters. If not specified, no scheduler will be used.
            e.g. lr_scheduler_params = { 'name': 'linear', 'num_warmup_steps': 0, 'num_training_steps': 5 }
        """
        train_input_data.set_labels(train_labels)
        train_data_loader = DataLoader(train_input_data, batch_size=self.config["batch_size"])
     
        if validation_input_data is not None and validation_labels is not None:
            validation_input_data.set_labels(validation_labels)
            validation_data_loader = DataLoader(validation_input_data, batch_size=self.config["batch_size"])

        self.to(self.config["device"])
        self.model.train()
        self.fine_tuning_head.train()
        optimizer = optimizer(self.parameters(), **optimizer_params)

        lr_scheduler = None
        if lr_scheduler_params is not None: 
            lr_scheduler = get_scheduler(optimizer=optimizer, **lr_scheduler_params)

        logger.info("Starting Fine-Tuning")
        for i in range(epochs):
            batch_loss = 0.0
            batches_processed = 0
            training_loop = tqdm(train_data_loader)
            for input_data, labels in training_loop:
                input_data = input_data.to(self.config["device"])
                labels = labels.to(self.config["device"])
                optimizer.zero_grad()
                output = self._forward(input_data)
                loss = loss_function(output, labels)
                batch_loss += loss.item()
                loss.backward()
                batches_processed += 1
                optimizer.step()
                
                training_loop.set_postfix({"loss": batch_loss/batches_processed})
                training_loop.set_description(f"Fine-Tuning: epoch {i+1}/{epochs}")
            
            if lr_scheduler is not None:
                lr_scheduler.step()

            if validation_input_data is not None and validation_labels is not None:
                with torch.no_grad():
                    validation_batches_processed = 0
                    val_loss = 0.0
                    validation_loop = tqdm(validation_data_loader, desc="Fine-Tuning Validation")
                    for input_data, val_labels in validation_loop:
                        input_data = input_data.to(self.config["device"])
                        val_labels = val_labels.to(self.config["device"])
                        output = self._forward(input_data)
                        validation_batches_processed += 1
                        val_loss += loss_function(output, val_labels).item()
                        validation_loop.set_postfix({"val_loss": val_loss/validation_batches_processed})
        logger.info(f"Fine-Tuning Complete. Epochs: {epochs}")
            
    def get_outputs(
            self, 
            input_data: HyenaDNADataset) -> np.ndarray:
        """Get the outputs of the fine-tuned model.
        
        Parameters
        ----------
        input_data : HyenaDNADataset
            The input data to get the outputs for.

        Returns
        -------
        np.ndarray
            The outputs of the model
        """
        data_loader = DataLoader(input_data, batch_size=self.config["batch_size"])

        self.to(self.config["device"])
        self.model.eval()
        self.fine_tuning_head.eval()

        batch_loop = tqdm(data_loader)
        outputs = []
        for batch, *labels in batch_loop:
            batch.to(self.config["device"])
            output = self._forward(batch)
            outputs.append(output.detach().cpu().numpy())
        
        return np.vstack(outputs)


from typing import Sequence
from helical.models.scgpt.data_sampler import SubsetSequentialSampler, SubsetsBatchSampler
from helical.models.scgpt.utils import set_seed

set_seed(1)


def _check_reorder(a: Sequence[int], b: Sequence[int]) -> bool:
    """Check if a is a shuffled version of b."""
    if len(a) != len(b):
        return False
    if len(a) == 0:
        return True
    if len(a) == 1:
        return a[0] == b[0]

    # Find the first element in a that is also in b
    first = a[0]
    for i in range(len(b)):
        if b[i] == first:
            break
    else:
        return False

    # Check if the rest of a is a shuffled version of b[:i] + b[i+1:]
    return _check_reorder(a[1:], b[:i] + b[i + 1 :])


def test_subset_sequential_sampler() -> None:
    """Test SubsetSequentialSampler."""
    # Test with a list of indices
    sampler = SubsetSequentialSampler([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    assert list(sampler) == list(range(10))


def test_subsets_batch_sampler() -> None:
    """Test SubsetsBatchSampler."""
    # Test with a list of subsets of indices
    subsets = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]

    # Test sequential sampling
    sampler = SubsetsBatchSampler(
        subsets,
        batch_size=3,
        intra_subset_shuffle=False,
        inter_subset_shuffle=False,
    )
    assert list(sampler) == [[0, 1, 2], [3, 4], [5, 6, 7], [8, 9]]

    # Test sequential sampling with drop_last
    sampler = SubsetsBatchSampler(
        subsets,
        batch_size=3,
        intra_subset_shuffle=False,
        inter_subset_shuffle=False,
        drop_last=True,
    )
    assert list(sampler) == [[0, 1, 2], [5, 6, 7]]

    # Test sampling with inter subset shuffle
    for _ in range(5):  # Run 5 times to check for consistent behavior
        sampler = SubsetsBatchSampler(
            subsets,
            batch_size=3,
            intra_subset_shuffle=False,
            inter_subset_shuffle=True,
        )
        sampled_idx = list(sampler)
        assert len(sampled_idx) == 4
        print(sampled_idx)
        if sampled_idx != [[0, 1, 2], [3, 4], [5, 6, 7], [8, 9]]:
            break
    else:
        assert False, "Inter-subset shuffle did not produce different orders in 5 attempts"
        assert _check_reorder(sampled_idx, [[0, 1, 2], [3, 4], [5, 6, 7], [8, 9]])

    # Test sampling with intra subset shuffle
    sampler = SubsetsBatchSampler(
        subsets,
        batch_size=3,
        intra_subset_shuffle=True,
        inter_subset_shuffle=False,
    )
    sampled_idx = list(sampler)
    assert len(sampled_idx) == 4
    sampled_idx_subset0 = sampled_idx[0] + sampled_idx[1]
    sampled_idx_subset1 = sampled_idx[2] + sampled_idx[3]
    assert sampled_idx_subset0 != [0, 1, 2, 3, 4]
    assert sampled_idx_subset1 != [5, 6, 7, 8, 9]
    assert _check_reorder(sampled_idx_subset0, [0, 1, 2, 3, 4])
    assert _check_reorder(sampled_idx_subset1, [5, 6, 7, 8, 9])

    # Test sampling with both intra and inter subset shuffle
    sampler = SubsetsBatchSampler(
        subsets,
        batch_size=3,
        intra_subset_shuffle=True,
        inter_subset_shuffle=True,
    )
    sampled_idx = list(sampler)
    assert len(sampled_idx) == 4
    sampled_idx_flatten = [item for sublist in sampled_idx for item in sublist]
    assert sampled_idx_flatten != [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    assert _check_reorder(sampled_idx_flatten, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])



import logging
import sys

logger = logging.getLogger("scGPT")
# check if logger has been initialized
if not logger.hasHandlers() or len(logger.handlers) == 0:
    logger.propagate = False
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(name)s - %(levelname)s - %(message)s", datefmt="%H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    
from .model import scGPT,scGPTConfig
from . import model_dir, tokenizer, scbank, utils
from .data_collator import DataCollator
from .dataset import Dataset
from .gene_embedding import GeneEmbedding
from .data_sampler import SubsetsBatchSampler
from .trainer import (
    prepare_data,
    prepare_dataloader,
    train,
    define_wandb_metrcis,
    evaluate,
    eval_testdata,
    test,
)
from .fine_tuning_model import scGPTFineTuningModel



import pickle as pkl
import requests
import json
import pickle as pkl
# imports
import logging
import torch
from tqdm.auto import trange
import re
import torch
from transformers import (
    BertForMaskedLM
)


logger = logging.getLogger(__name__)

def load_mappings(gene_symbols):
    server = "https://rest.ensembl.org"
    ext = "/lookup/symbol/homo_sapiens"
    headers={ "Content-Type" : "application/json", "Accept" : "application/json"}

    # r = requests.post(server+ext, headers=headers, data='{ "symbols" : ["BRCA2", "BRAF" ] }')

    gene_id_to_ensemble = dict()

    for i in range(0, len(gene_symbols), 1000):
        # print(i+1000,"/",len(test.var['gene_symbols']))
        symbols = {'symbols':gene_symbols[i:i+1000].tolist()}
        r = requests.post(server+ext, headers=headers, data=json.dumps(symbols))
        decoded = r.json()
        gene_id_to_ensemble.update(decoded)
        # print(repr(decoded))

    pkl.dump(gene_id_to_ensemble, open('./human_gene_to_ensemble_id.pkl', 'wb'))
    return gene_id_to_ensemble


# extract embeddings
def get_embs(
    model,
    filtered_input_data,
    emb_mode,
    layer_to_quant,
    pad_token_id,
    forward_batch_size,
    gene_token_dict,
    device,
    silent=False,
    
):
    model_input_size = get_model_input_size(model)
    total_batch_length = len(filtered_input_data)
    embs_list = []

    #  Check if CLS and EOS token is present in the token dictionary
    cls_present = any("<cls>" in key for key in gene_token_dict.keys())
    eos_present = any("<eos>" in key for key in gene_token_dict.keys())
    if emb_mode == "cls":
        assert cls_present, "<cls> token missing in token dictionary"
        # Check to make sure that the first token of the filtered input data is cls token
        cls_token_id = gene_token_dict["<cls>"]
        assert (
            filtered_input_data["input_ids"][0][0] == cls_token_id
        ), "First token is not <cls> token value"
    elif emb_mode == "cell":
        if cls_present:
            logger.warning(
                "CLS token present in token dictionary, excluding from average."
            )
        if eos_present:
            logger.warning(
                "EOS token present in token dictionary, excluding from average."
            )

    overall_max_len = 0
    for i in trange(0, total_batch_length, forward_batch_size, leave=(not silent)):
        max_range = min(i + forward_batch_size, total_batch_length)

        minibatch = filtered_input_data.select([i for i in range(i, max_range)])

        max_len = int(max(minibatch["length"]))
        original_lens = torch.tensor(minibatch["length"],device=device)
        minibatch.set_format(type="torch",device=device)

        input_data_minibatch = minibatch["input_ids"]
        input_data_minibatch = pad_tensor_list(
            input_data_minibatch, max_len, pad_token_id, model_input_size
        ).to(device)

        model = model.to(device)
        with torch.no_grad():
            outputs = model(
                input_ids=input_data_minibatch,
                attention_mask=gen_attention_mask(minibatch),
            )

        embs_i = outputs.hidden_states[layer_to_quant]

        if emb_mode == "cell":
            if cls_present:
                non_cls_embs = embs_i[:, 1:, :]  # Get all layers except the cls embs
                if eos_present:
                    mean_embs = mean_nonpadding_embs(non_cls_embs, original_lens - 2)
                else:
                    mean_embs = mean_nonpadding_embs(non_cls_embs, original_lens - 1)
            else:
                mean_embs = mean_nonpadding_embs(embs_i, original_lens)
            
            embs_list.append(mean_embs)
            del mean_embs

        elif emb_mode == "gene":
                embs_list.append(embs_i)
        
        elif emb_mode == "cls":
            cls_embs = embs_i[:, 0, :].clone().detach()  # CLS token layer
            embs_list.append(cls_embs)
            del cls_embs

        overall_max_len = max(overall_max_len, max_len)
        del outputs
        del minibatch
        del input_data_minibatch
        del embs_i

        torch.cuda.empty_cache()

    if emb_mode == "cell" or emb_mode == "cls":
        embs_stack = torch.cat(embs_list, dim=0)
    elif emb_mode == "gene":
        embs_stack = pad_tensor_list(
            embs_list,
            overall_max_len,
            pad_token_id,
            model_input_size,
            1,
            pad_3d_tensor,
        )
    return embs_stack


def downsample_and_sort(data, max_ncells):
    num_cells = len(data)
    # if max number of cells is defined, then shuffle and subsample to this max number
    if max_ncells is not None:
        if num_cells > max_ncells:
            data = data.shuffle(seed=42)
            num_cells = max_ncells
    data_subset = data.select([i for i in range(num_cells)])
    # sort dataset with largest cell first to encounter any memory errors earlier
    data_sorted = data_subset.sort("length", reverse=True)
    return data_sorted



def quant_layers(model):
    layer_nums = []
    for name, parameter in model.named_parameters():
        if "layer" in name:
            layer_nums += [int(name.split("layer.")[1].split(".")[0])]
    return int(max(layer_nums)) + 1


def get_model_input_size(model):
    return int(re.split("\(|,", str(model.bert.embeddings.position_embeddings))[1])

def load_model(model_type, model_directory, device):
    if model_type == "Pretrained":
        model = BertForMaskedLM.from_pretrained(
            model_directory, output_hidden_states=True, output_attentions=False
        )
    # put the model in eval mode for fwd pass and load onto the GPU if available
    model.eval()
    model = model.to(device)
    return model

def pad_tensor(tensor, pad_token_id, max_len):
    tensor = torch.nn.functional.pad(
        tensor, pad=(0, max_len - tensor.numel()), mode="constant", value=pad_token_id
    )

    return tensor

def pad_3d_tensor(tensor, pad_token_id, max_len, dim):
    if dim == 0:
        raise Exception("dim 0 usually does not need to be padded.")
    if dim == 1:
        pad = (0, 0, 0, max_len - tensor.size()[dim])
    elif dim == 2:
        pad = (0, max_len - tensor.size()[dim], 0, 0)
    tensor = torch.nn.functional.pad(
        tensor, pad=pad, mode="constant", value=pad_token_id
    )
    return tensor

# pad list of tensors and convert to tensor
def pad_tensor_list(
    tensor_list,
    dynamic_or_constant,
    pad_token_id,
    model_input_size,
    dim=None,
    padding_func=None,
):
    # determine maximum tensor length
    if dynamic_or_constant == "dynamic":
        max_len = max([tensor.squeeze().numel() for tensor in tensor_list])
    elif isinstance(dynamic_or_constant, int):
        max_len = dynamic_or_constant
    else:
        max_len = model_input_size
        logger.warning(
            "If padding style is constant, must provide integer value. "
            f"Setting padding to max input size {model_input_size}."
        )

    # pad all tensors to maximum length
    if dim is None:
        tensor_list = [
            pad_tensor(tensor, pad_token_id, max_len) for tensor in tensor_list
        ]
    else:
        tensor_list = [
            padding_func(tensor, pad_token_id, max_len, dim) for tensor in tensor_list
        ]
    # return stacked tensors
    if padding_func != pad_3d_tensor:
        return torch.stack(tensor_list)
    else:
        return torch.cat(tensor_list, 0)


def gen_attention_mask(minibatch_encoding, max_len=None):
    if max_len is None:
        max_len = max(minibatch_encoding["length"])
    original_lens = minibatch_encoding["length"]
    attention_mask = [
        [1] * original_len + [0] * (max_len - original_len)
        if original_len <= max_len
        else [1] * max_len
        for original_len in original_lens
    ]
    return torch.tensor(attention_mask, device=minibatch_encoding["length"].device)


# get cell embeddings excluding padding
def mean_nonpadding_embs(embs, original_lens, dim=1):
    # create a mask tensor based on padding lengths
    mask = torch.arange(embs.size(dim), device=embs.device) < original_lens.unsqueeze(1)
    if embs.dim() == 3:
        # fill the masked positions in embs with zeros
        masked_embs = embs.masked_fill(~mask.unsqueeze(2), 0.0)

        # compute the mean across the non-padding dimensions
        mean_embs = masked_embs.sum(dim) / original_lens.view(-1, 1).float()

    elif embs.dim() == 2:
        masked_embs = embs.masked_fill(~mask, 0.0)
        mean_embs = masked_embs.sum(dim) / original_lens.float()
    return mean_embs



import os
import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
import time
import traceback
import numpy as np

from anndata import AnnData
import scanpy as sc
from typing import List, Tuple, Dict, Optional
from helical.models.scgpt.tokenizer import tokenize_and_pad_batch, random_mask_value
from helical.models.scgpt import SubsetsBatchSampler
from helical.models.scgpt.loss import (
    masked_relative_error,
    criterion_neg_log_bernoulli,
)
from helical.models.scgpt.utils import eval_scib_metrics
import warnings
from scipy.sparse import issparse
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score


def prepare_data(
    tokenized_train,
    tokenized_valid,
    train_batch_labels,
    valid_batch_labels,
    config,
    epoch,
    train_celltype_labels=None,
    valid_celltype_labels=None,
    sort_seq_batch=False,
) -> Tuple[Dict[str, torch.Tensor]]:
    assert config.task in ["annotation", "integration", "perturb", "multiomic"]
    masked_values_train = random_mask_value(
        tokenized_train["values"],
        mask_ratio=config.mask_ratio,
        mask_value=config.mask_value,
        pad_value=config.pad_value,
    )
    masked_values_valid = random_mask_value(
        tokenized_valid["values"],
        mask_ratio=config.mask_ratio,
        mask_value=config.mask_value,
        pad_value=config.pad_value,
    )
    print(
        f"random masking at epoch {epoch:3d}, ratio of masked values in train: ",
        f"{(masked_values_train == config.mask_value).sum() / (masked_values_train - config.pad_value).count_nonzero():.4f}",
    )

    input_gene_ids_train, input_gene_ids_valid = (
        tokenized_train["genes"],
        tokenized_valid["genes"],
    )
    input_values_train, input_values_valid = masked_values_train, masked_values_valid

    target_values_train, target_values_valid = (
        tokenized_train["values"],
        tokenized_valid["values"],
    )

    tensor_batch_labels_train = torch.from_numpy(train_batch_labels).long()
    tensor_batch_labels_valid = torch.from_numpy(valid_batch_labels).long()

    if config.task == "annotation":
        tensor_celltype_labels_train = torch.from_numpy(train_celltype_labels).long()
        tensor_celltype_labels_valid = torch.from_numpy(valid_celltype_labels).long()

    if config.task == "multiomic":
        tensor_mod_types_train, tensor_mod_types_valid = (
            tokenized_train["mod_types"].long(),
            tokenized_valid["mod_types"].long(),
        )

    if sort_seq_batch:
        train_sort_ids = np.argsort(train_batch_labels)
        input_gene_ids_train = input_gene_ids_train[train_sort_ids]
        input_values_train = input_values_train[train_sort_ids]
        target_values_train = target_values_train[train_sort_ids]
        tensor_batch_labels_train = tensor_batch_labels_train[train_sort_ids]
        if config.task == "annotation":
            tensor_celltype_labels_train = tensor_celltype_labels_train[train_sort_ids]
        if config.task == "multiomic":
            tensor_mod_types_train = tensor_mod_types_train[train_sort_ids]
        valid_sort_ids = np.argsort(valid_batch_labels)
        input_gene_ids_valid = input_gene_ids_valid[valid_sort_ids]
        input_values_valid = input_values_valid[valid_sort_ids]
        target_values_valid = target_values_valid[valid_sort_ids]
        tensor_batch_labels_valid = tensor_batch_labels_valid[valid_sort_ids]
        if config.task == "annotation":
            tensor_celltype_labels_valid = tensor_celltype_labels_valid[valid_sort_ids]
        if config.task == "multiomic":
            tensor_mod_types_valid = tensor_mod_types_valid[valid_sort_ids]

    train_data_pt = {
        "gene_ids": input_gene_ids_train,
        "values": input_values_train,
        "target_values": target_values_train,
        "batch_labels": tensor_batch_labels_train,
    }
    valid_data_pt = {
        "gene_ids": input_gene_ids_valid,
        "values": input_values_valid,
        "target_values": target_values_valid,
        "batch_labels": tensor_batch_labels_valid,
    }
    if config.task == "annotation":
        train_data_pt["celltype_labels"] = tensor_celltype_labels_train
        valid_data_pt["celltype_labels"] = tensor_celltype_labels_valid
    if config.task == "multiomic":
        train_data_pt["mod_types"] = tensor_mod_types_train
        valid_data_pt["mod_types"] = tensor_mod_types_valid
    return train_data_pt, valid_data_pt


class SeqDataset(Dataset):
    def __init__(self, data: Dict[str, torch.Tensor]):
        self.data = data

    def __len__(self):
        return self.data["gene_ids"].shape[0]

    def __getitem__(self, idx):
        return {k: v[idx] for k, v in self.data.items()}


# data_loader
def prepare_dataloader(
    data_pt: Dict[str, torch.Tensor],
    batch_size: int,
    shuffle: bool = False,
    intra_domain_shuffle: bool = False,
    drop_last: bool = False,
    num_workers: int = 0,
    per_seq_batch_sample: bool = False,
) -> DataLoader:
    dataset = SeqDataset(data_pt)

    if per_seq_batch_sample:
        # find the indices of samples in each seq batch
        subsets = []
        batch_labels_array = data_pt["batch_labels"].numpy()
        for batch_label in np.unique(batch_labels_array):
            batch_indices = np.where(batch_labels_array == batch_label)[0].tolist()
            subsets.append(batch_indices)
        data_loader = DataLoader(
            dataset=dataset,
            batch_sampler=SubsetsBatchSampler(
                subsets,
                batch_size,
                intra_subset_shuffle=intra_domain_shuffle,
                inter_subset_shuffle=shuffle,
                drop_last=drop_last,
            ),
            num_workers=num_workers,
            pin_memory=True,
        )
        return data_loader

    data_loader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        num_workers=num_workers,
        pin_memory=True,
    )
    return data_loader


def train(
    model: nn.Module,
    loader: DataLoader,
    vocab,
    criterion_gep_gepc,
    criterion_dab,
    criterion_cls,
    scaler,
    optimizer,
    scheduler,
    device,
    config,
    logger,
    epoch,
) -> None:
    """
    Train the model for one epoch.
    """
    import wandb

    total_loss, total_gep, total_cls, total_gepc, total_ecs, total_dab = (
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    )
    total_zero_log_prob, total_gepc_zero_log_prob = 0.0, 0.0
    # total_error = 0.0
    log_interval = config.log_interval
    start_time = time.time()

    num_batches = len(loader)
    for batch, batch_data in enumerate(loader):
        input_gene_ids = batch_data["gene_ids"].to(device)
        input_values = batch_data["values"].to(device)
        target_values = batch_data["target_values"].to(device)
        batch_labels = batch_data["batch_labels"].to(device)
        if config.task == "annotation":
            celltype_labels = batch_data["celltype_labels"].to(device)
        if config.task == "multiomic":
            mod_types = batch_data["mod_types"].to(device)

        src_key_padding_mask = input_gene_ids.eq(vocab[config.pad_token])

        with torch.cuda.amp.autocast(enabled=config.amp):
            output_dict = model(
                input_gene_ids,
                input_values,
                src_key_padding_mask=src_key_padding_mask,
                batch_labels=batch_labels
                if config.use_batch_labels or config.DSBN
                else None,
                CLS=config.CLS,
                MVC=config.GEPC,
                ECS=config.ESC,
                mod_types=mod_types if config.use_mod else None,
                # do_sample=do_sample_in_train,
                # generative_training=False
            )

            masked_positions = input_values.eq(
                config.mask_value
            )  # the postions to predict
            loss = 0.0
            metrics_to_log = {}
            if config.GEP:
                loss_gep = criterion_gep_gepc(
                    output_dict["mlm_output"], target_values, masked_positions
                )
                loss = loss + loss_gep
                metrics_to_log = {"train/gep": loss_gep.item()}
            if config.GEP and config.explicit_zero_prob:
                loss_zero_log_prob = criterion_neg_log_bernoulli(
                    output_dict["mlm_zero_probs"], target_values, masked_positions
                )
                loss = loss + loss_zero_log_prob
                metrics_to_log.update({"train/nzlp": loss_zero_log_prob.item()})

            if config.GEPC:
                loss_gepc = criterion_gep_gepc(
                    output_dict["mvc_output"], target_values, masked_positions
                )
                loss = loss + loss_gepc
                metrics_to_log.update({"train/mvc": loss_gepc.item()})

            if config.GEPC and config.explicit_zero_prob:
                loss_gepc_zero_log_prob = criterion_neg_log_bernoulli(
                    output_dict["mvc_zero_probs"], target_values, masked_positions
                )
                loss = loss + loss_gepc_zero_log_prob
                metrics_to_log.update(
                    {"train/mvc_nzlp": loss_gepc_zero_log_prob.item()}
                )

            if config.CLS:
                loss_cls = criterion_cls(output_dict["cls_output"], celltype_labels)
                loss = loss + loss_cls
                metrics_to_log.update({"train/cls": loss_cls.item()})

                error_rate = 1 - (
                    (output_dict["cls_output"].argmax(1) == celltype_labels)
                    .sum()
                    .item()
                ) / celltype_labels.size(0)

            if config.ESC:
                loss_ecs = 10 * output_dict["loss_ecs"]
                loss = loss + loss_ecs
                metrics_to_log.update({"train/ecs": loss_ecs.item()})

            if config.DAR:
                # try weighting and separate optimizer
                loss_dab = criterion_dab(output_dict["dab_output"], batch_labels)
                loss = loss + config.dab_weight * loss_dab
                metrics_to_log.update({"train/dab": loss_dab.item()})

        model.zero_grad()
        scaler.scale(loss).backward()
        scaler.unscale_(optimizer)

        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings("always")
            torch.nn.utils.clip_grad_norm_(
                model.parameters(),
                1.0,
                error_if_nonfinite=False if scaler.is_enabled() else True,
            )
            if len(w) > 0:
                logger.warning(
                    f"Found infinite gradient. This may be caused by the gradient "
                    f"scaler. The current scale is {scaler.get_scale()}. This warning "
                    "can be ignored if no longer occurs after autoscaling of the scaler."
                )
        scaler.step(optimizer)
        scaler.update()

        wandb.log(metrics_to_log)

        total_loss += loss.item()
        total_gep += loss_gep.item() if config.GEP else 0.0
        total_cls += loss_cls.item() if config.CLS else 0.0
        total_gepc += loss_gepc.item() if config.GEPC else 0.0
        total_ecs += loss_ecs.item() if config.ESC else 0.0
        total_dab += loss_dab.item() if config.DAR else 0.0
        total_zero_log_prob += (
            loss_zero_log_prob.item()
            if config.GEP and config.explicit_zero_prob
            else 0.0
        )
        total_gepc_zero_log_prob += (
            loss_gepc_zero_log_prob.item()
            if config.GEPC and config.explicit_zero_prob
            else 0.0
        )
        if batch % log_interval == 0 and batch > 0:
            lr = scheduler.get_last_lr()[0]
            ms_per_batch = (time.time() - start_time) * 1000 / log_interval
            cur_loss = total_loss / log_interval
            cur_gep = total_gep / log_interval if config.GEP else 0.0
            cur_cls = total_cls / log_interval if config.CLS else 0.0
            cur_gepc = total_gepc / log_interval if config.GEPC else 0.0
            cur_ecs = total_ecs / log_interval if config.ESC else 0.0
            cur_dab = total_dab / log_interval if config.DAR else 0.0
            cur_zero_log_prob = (
                total_zero_log_prob / log_interval if config.explicit_zero_prob else 0.0
            )
            cur_gepc_zero_log_prob = (
                total_gepc_zero_log_prob / log_interval
                if config.GEPC and config.explicit_zero_prob
                else 0.0
            )
            # cur_error = total_error / log_interval

            logger.info(
                f"| epoch {epoch:3d} | {batch:3d}/{num_batches:3d} batches | "
                f"lr {lr:05.5f} | ms/batch {ms_per_batch:5.2f} | "
                f"loss {cur_loss:5.2f} | "
                + (f"gep {cur_gep:5.2f} |" if config.GEP else "")
                + (f"cls {cur_cls:5.2f} | " if config.CLS else "")
                # + (f"err {cur_error:5.2f} | " if config.CLS else "")
                + (f"gepc {cur_gepc:5.2f} |" if config.GEPC else "")
                + (f"ecs {cur_ecs:5.2f} |" if config.ESC else "")
                + (f"dar {cur_dab:5.2f} |" if config.DAR else "")
            )
            total_loss = 0
            total_gep = 0
            total_cls = 0
            total_gepc = 0
            total_ecs = 0
            total_dab = 0
            total_zero_log_prob = 0
            total_gepc_zero_log_prob = 0
            # total_error = 0
            start_time = time.time()


def define_wandb_metrcis():
    import wandb

    wandb.define_metric("valid/loss", summary="min", step_metric="epoch")
    wandb.define_metric("test/avg_bio", summary="max")


def evaluate(
    model: nn.Module,
    loader: DataLoader,
    vocab,
    criterion_gep_gepc,
    criterion_dab,
    criterion_cls,
    device,
    config,
    epoch,
) -> float:
    """
    Evaluate the model on the evaluation data.
    """
    import wandb

    model.eval()
    total_loss = 0.0
    # total_error = 0.0
    total_dab = 0.0
    total_num = 0
    with torch.no_grad():
        for batch_data in loader:
            input_gene_ids = batch_data["gene_ids"].to(device)
            input_values = batch_data["values"].to(device)
            target_values = batch_data["target_values"].to(device)
            batch_labels = batch_data["batch_labels"].to(device)
            if config.task == "annotation":
                celltype_labels = batch_data["celltype_labels"].to(device)
            if config.task == "multiomic":
                mod_types = batch_data["mod_types"].to(device)

            src_key_padding_mask = input_gene_ids.eq(vocab[config.pad_token])

            with torch.cuda.amp.autocast(enabled=config.amp):
                output_dict = model(
                    input_gene_ids,
                    input_values,
                    src_key_padding_mask=src_key_padding_mask,
                    batch_labels=batch_labels
                    if config.use_batch_labels or config.DSBN
                    else None,
                    CLS=config.CLS,  # evaluation does not need CLS or CCE
                    MVC=False,
                    ECS=False,
                    mod_types=mod_types if config.use_mod else None,
                    # do_sample=do_sample_in_train,
                    # generative_training = False,
                )
                if config.task == "annotation":
                    output_values = output_dict["cls_output"]
                    loss = criterion_cls(output_values, celltype_labels)

                elif config.task in ["integration", "multiomic"]:
                    output_values = output_dict["mlm_output"]
                    masked_positions = input_values.eq(config.mask_value)
                    loss = criterion_gep_gepc(
                        output_values, target_values, masked_positions
                    )

                if config.DAR:
                    loss_dab = criterion_dab(output_dict["dab_output"], batch_labels)

            total_loss += loss.item() * len(input_gene_ids)

            if config.DAR:
                total_dab += (
                    loss_dab.item() * len(input_gene_ids) if config.DAR else 0.0
                )
            else:
                total_dab = 0

            total_num += len(input_gene_ids)

    wandb.log(
        {
            "valid/loss": (total_loss + config.dab_weight * total_dab) / total_num,
            "epoch": epoch,
        },
    )

    return total_loss / total_num


def predict(
    model: nn.Module,
    loader: DataLoader,
    vocab,
    config,
    device,
) -> float:
    """
    Evaluate the model on the evaluation data.
    """
    model.eval()

    predictions = []
    with torch.no_grad():
        for batch_data in loader:
            input_gene_ids = batch_data["gene_ids"].to(device)
            input_values = batch_data["values"].to(device)
            target_values = batch_data["target_values"].to(device)
            batch_labels = batch_data["batch_labels"].to(device)
            celltype_labels = batch_data["celltype_labels"].to(device)

            src_key_padding_mask = input_gene_ids.eq(vocab[config.pad_token])

            with torch.cuda.amp.autocast(enabled=config.amp):
                output_dict = model(
                    input_gene_ids,
                    input_values,
                    src_key_padding_mask=src_key_padding_mask,
                    batch_labels=batch_labels
                    if config.use_batch_labels or config.DSBN
                    else None,
                    CLS=config.CLS,
                    MVC=config.GEPC,
                    ECS=config.ESC,
                )

                output_values = output_dict["cls_output"]
                preds = output_values.argmax(1).cpu().numpy()
                predictions.append(preds)

    return np.concatenate(predictions, axis=0)


# %% inference
def test(
    model: nn.Module, adata: DataLoader, gene_ids, vocab, config, device, logger
) -> float:
    all_counts = (
        adata.layers[config.input_layer_key].A
        if issparse(adata.layers[config.input_layer_key])
        else adata.layers[config.input_layer_key]
    )

    celltypes_labels = adata.obs["celltype_id"].tolist()  # make sure count from 0
    celltypes_labels = np.array(celltypes_labels)

    batch_ids = adata.obs["batch_id"].tolist()
    batch_ids = np.array(batch_ids)

    tokenized_test = tokenize_and_pad_batch(
        all_counts,
        gene_ids,
        max_len=config.max_seq_len,
        vocab=vocab,
        pad_token=config.pad_token,
        pad_value=config.pad_value,
        append_cls=True,  # append <cls> token at the beginning
        include_zero_gene=config.include_zero_gene,
    )

    input_values_test = random_mask_value(
        tokenized_test["values"],
        mask_ratio=config.mask_ratio,
        mask_value=config.mask_value,
        pad_value=config.pad_value,
    )

    test_data_pt = {
        "gene_ids": tokenized_test["genes"],
        "values": input_values_test,
        "target_values": tokenized_test["values"],
        "batch_labels": torch.from_numpy(batch_ids).long(),
        "celltype_labels": torch.from_numpy(celltypes_labels).long(),
    }

    test_loader = DataLoader(
        dataset=SeqDataset(test_data_pt),
        batch_size=config.batch_size,
        shuffle=False,
        drop_last=False,
        num_workers=min(len(os.sched_getaffinity(0)), config.batch_size // 2),
        pin_memory=True,
    )

    model.eval()
    predictions = predict(
        model,
        test_loader,
        vocab,
        config,
        device,
    )

    # compute accuracy, precision, recall, f1
    accuracy = accuracy_score(celltypes_labels, predictions)
    precision = precision_score(celltypes_labels, predictions, average="macro")
    recall = recall_score(celltypes_labels, predictions, average="macro")
    macro_f1 = f1_score(celltypes_labels, predictions, average="macro")
    micro_f1 = f1_score(celltypes_labels, predictions, average="micro")

    logger.info(
        f"Accuracy: {accuracy:.3f}, Precision: {precision:.3f}, Recall: {recall:.3f}, "
        f"Macro F1: {macro_f1:.3f}, Micro F1: {micro_f1:.3f}"
    )

    results = {
        "test/accuracy": accuracy,
        "test/precision": precision,
        "test/recall": recall,
        "test/macro_f1": macro_f1,
        "test/micro_f1": micro_f1,
    }

    return predictions, celltypes_labels, results


def eval_testdata(
    model: nn.Module,
    adata_t: AnnData,
    gene_ids,
    vocab,
    config,
    logger,
    include_types: List[str] = ["cls"],
) -> Optional[Dict]:
    """evaluate the model on test dataset of adata_t"""
    model.eval()

    # copy adata_t to avoid reuse previously computed results stored in adata_t
    adata_t = adata_t.copy()

    all_counts = (
        adata_t.layers[config.input_layer_key].A
        if issparse(adata_t.layers[config.input_layer_key])
        else adata_t.layers[config.input_layer_key]
    )

    celltypes_labels = adata_t.obs["celltype"].tolist()
    celltypes_labels = np.array(celltypes_labels)

    batch_ids = adata_t.obs["batch_id"].tolist()
    batch_ids = np.array(batch_ids)

    # Evaluate cls cell embeddings
    if "cls" in include_types:
        logger.info("Evaluating cls cell embeddings")
        tokenized_all = tokenize_and_pad_batch(
            all_counts,
            gene_ids,
            max_len=config.max_seq_len,
            vocab=vocab,
            pad_token=config.pad_token,
            pad_value=config.pad_value,
            append_cls=True,  # append <cls> token at the beginning
            include_zero_gene=config.include_zero_gene,
        )
        all_gene_ids, all_values = tokenized_all["genes"], tokenized_all["values"]
        src_key_padding_mask = all_gene_ids.eq(vocab[config.pad_token])
        with torch.no_grad(), torch.cuda.amp.autocast(enabled=config.amp):
            cell_embeddings = model.encode_batch(
                all_gene_ids,
                all_values.float(),
                src_key_padding_mask=src_key_padding_mask,
                batch_size=config.batch_size,
                batch_labels=torch.from_numpy(batch_ids).long()
                if config.DSBN or config.DAR or config.use_batch_labels
                else None,
                time_step=0,
                return_np=True,
            )
        cell_embeddings = cell_embeddings / np.linalg.norm(
            cell_embeddings, axis=1, keepdims=True
        )

        adata_t.obsm["X_scGPT"] = cell_embeddings

        results = {}
        try:
            results = eval_scib_metrics(adata_t)
        except Exception as e:
            traceback.print_exc()
            logger.error(e)

        sc.pp.neighbors(adata_t, use_rep="X_scGPT")
        sc.tl.umap(adata_t, min_dist=0.3)
        fig = sc.pl.umap(
            adata_t,
            color=["str_batch"],
            title=[f"batch, avg_bio = {results.get('avg_bio', 0.0):.4f}"],
            frameon=False,
            return_fig=True,
            show=False,
        )

        results["batch_umap"] = fig

        sc.pp.neighbors(adata_t, use_rep="X_scGPT")
        sc.tl.umap(adata_t, min_dist=0.3)
        fig = sc.pl.umap(
            adata_t,
            color=["celltype"],
            title=[
                f"celltype, avg_bio = {results.get('avg_bio', 0.0):.4f}",
            ],
            frameon=False,
            return_fig=True,
            show=False,
        )

        results["celltype_umap"] = fig

    if len(include_types) == 1:
        return results



from torch.utils.data import Dataset
import torch

class HelixmRNADataset(Dataset):
    def __init__(self, sequences, tokenizer):
        self.sequences = sequences
        self.tokenizer = tokenizer
        self.max_length = len(max(sequences, key=len))+10
        self.labels = None

    def set_labels(self, labels):
        self.labels = labels

    def __len__(self):
        return len(self.sequences)

    def __getitem__(self, idx):
        seqs = self.sequences[idx]
        
        # Tokenize all sequences in the batch
        encoded = self.tokenizer(
            seqs,
            truncation=True,
            padding="max_length",
            max_length=self.max_length,
            return_special_tokens_mask=True,
            return_tensors='pt'
        )
        
        # Prepare output dictionary
        output = {
            'input_ids': encoded['input_ids'].squeeze(0),
            'special_tokens_mask': encoded['special_tokens_mask'].squeeze(0),
        }
        
        # Add labels if they exist
        if self.labels is not None:
            output['labels'] = torch.tensor(self.labels[idx])

        return output


###############################################################################
# Auto-generated by `jupyter-book config`
# If you wish to continue using _config.yml, make edits to that file and
# re-generate this one.
###############################################################################
author = 'Helical Team'
autosummary_generate = True
bibtex_bibfiles = ['references.bib']
comments_config = {'hypothesis': False, 'utterances': False}
copyright = '2024'
exclude_patterns = ['**.ipynb_checkpoints', '.DS_Store', 'Thumbs.db', '_build']
extensions = ['sphinx_togglebutton', 'sphinx_copybutton', 'myst_nb', 'jupyter_book', 'sphinx_thebe', 'sphinx_comments', 'sphinx_external_toc', 'sphinx.ext.intersphinx', 'sphinx_design', 'sphinx_book_theme', 'sphinx_inline_tabs', 'sphinx_proof', 'sphinx_examples', 'hoverxref.extension', 'sphinx.ext.autodoc', 'sphinx.ext.viewcode', 'sphinx.ext.autosummary', 'sphinx.ext.napoleon', 'sphinxcontrib.bibtex', 'sphinx_jupyterbook_latex', 'sphinx_multitoc_numbering']
external_toc_exclude_missing = False
external_toc_path = '_toc.yml'
hoverxref_intersphinx = ['sphinxproof']
html_baseurl = ''
html_css_files = ['my_theme copy.css', 'my_theme.css']
html_favicon = 'docs/official_logo_fav.png'
html_logo = 'docs/official_logo_small.png'
html_sourcelink_suffix = ''
html_static_path = ['_static']
html_theme = 'sphinx_book_theme'
html_theme_options = {'search_bar_text': 'Search this book...', 'launch_buttons': {'notebook_interface': 'classic', 'binderhub_url': '', 'jupyterhub_url': '', 'thebe': False, 'colab_url': ''}, 'path_to_docs': 'docs', 'repository_url': 'https://github.com/helicalAI/helical', 'repository_branch': 'main', 'extra_footer': '', 'home_page_in_toc': True, 'announcement': '', 'analytics': {'google_analytics_id': ''}, 'use_repository_button': True, 'use_edit_page_button': False, 'use_issues_button': True}
html_title = 'Helical Documentation'
intersphinx_mapping = {'ebp': ['https://executablebooks.org/en/latest/', None], 'myst-parser': ['https://myst-parser.readthedocs.io/en/latest/', None], 'myst-nb': ['https://myst-nb.readthedocs.io/en/latest/', None], 'sphinx': ['https://www.sphinx-doc.org/en/master', None], 'nbformat': ['https://nbformat.readthedocs.io/en/latest', None], 'sd': ['https://sphinx-design.readthedocs.io/en/latest', None], 'sphinxproof': ['https://sphinx-proof.readthedocs.io/en/latest/', None]}
latex_engine = 'pdflatex'
mathjax3_config = {'tex': {'macros': {'N': '\\mathbb{N}', 'floor': ['\\lfloor#1\\rfloor', 1], 'bmat': ['\\left[\\begin{array}'], 'emat': ['\\end{array}\\right]']}}}
myst_enable_extensions = ['colon_fence', 'dollarmath', 'linkify', 'substitution', 'tasklist']
myst_url_schemes = ['mailto', 'http', 'https']
napoleon_include_init_with_doc = True
napoleon_numpy_docstring = True
nb_execution_allow_errors = False
nb_execution_cache_path = ''
nb_execution_excludepatterns = []
nb_execution_in_temp = False
nb_execution_mode = 'off'
nb_execution_timeout = 30
nb_output_stderr = 'show'
numfig = True
pygments_style = 'sphinx'
suppress_warnings = ['myst.domains']
use_jupyterbook_latex = True
use_multitoc_numbering = True



from typing import Literal, Optional
from helical.models.base_models import HelicalBaseFineTuningHead, HelicalBaseFineTuningModel
from helical.models.geneformer import Geneformer, GeneformerConfig
import torch
from torch import optim
from torch.nn.modules import loss
from helical.models.geneformer.geneformer_utils import gen_attention_mask, get_model_input_size, pad_tensor_list
from datasets import Dataset
import logging
from tqdm import trange
from transformers import get_scheduler


logger = logging.getLogger(__name__)

class GeneformerFineTuningModel(HelicalBaseFineTuningModel, Geneformer):
    """GeneformerFineTuningModel.
    
    Fine-tuning model for the Geneformer model.

    Example
    ----------
    ```python
    from helical import GeneformerConfig, GeneformerFineTuningModel

    # Prepare the data
    ann_data = ad.read_h5ad("dataset.h5ad")

    # Get the desired label class
    cell_types = list(ann_data.obs.cell_type)

    # Create a dictionary mapping the classes to unique integers for training
    label_set = set(cell_types)
    class_id_dict = dict(zip(label_set, [i for i in range(len(label_set))]))

    for i in range(len(cell_types)):
        cell_types[i] = class_id_dict[cell_types[i]]

    # Add this column to the Dataset
    dataset = dataset.add_column('cell_types', cell_types)

    # Create the fine-tuning model
    model_config = GeneformerConfig(model_name="gf-12L-95M-i4096", batch_size=10)
    geneformer_fine_tune = GeneformerFineTuningModel(
        geneformer_config=model_config, 
        fine_tuning_head="classification", 
        label="cell_types", 
        output_size=len(label_set)
    )

    # Process the data for training
    dataset = geneformer_fine_tune.process_data(ann_data)

    # Fine-tune
    geneformer_fine_tune.train(train_dataset=dataset)

    # Get outputs of the fine-tuned model
    outputs = geneformer_fine_tune.get_outputs(dataset)

    # Get the embeddings of the fine-tuned model
    embeddings = geneformer_fine_tune.get_embeddings(dataset)
    ```
    
    Parameters
    ----------
    geneformer_config : GeneformerConfig
        The Geneformer configs to fine-tune, the same as instantiating the standard Geneformer model.
    fine_tuning_head : Literal["classification", "regression"] | HelicalBaseFineTuningHead
        The fine-tuning head that is appended to the model. This can either be a string (options available: "classification", "regression") specifying the task or a custom fine-tuning head inheriting from HelicalBaseFineTuningHead.
    output_size : Optional[int]
        The output size of the fine-tuning model. This is required if the `fine_tuning_head` is a string specified task. For a classification task this is number of unique classes.

    Methods
    -------
    train(train_dataset: Dataset, optimizer: optim, optimizer_params: dict, loss_function: loss, label: str, epochs: int, freeze_layers: int, validation_dataset: Optional[Dataset], lr_scheduler_params: Optional[dict], silent = False)
        Fine-tunes the Geneformer model.
    get_outputs(dataset: Dataset, silent = False)
        Get outputs from the fine-tuned model on the given processed dataset.
    """
    def __init__(self,
                 geneformer_config: GeneformerConfig, 
                 fine_tuning_head: Literal["classification", "regression"] | HelicalBaseFineTuningHead, 
                 output_size: Optional[int]=None):
        
        HelicalBaseFineTuningModel.__init__(self, fine_tuning_head, output_size)
        Geneformer.__init__(self, geneformer_config)

        self.fine_tuning_head.set_dim_size(self.config["embsize"])

    def _forward(self, input_ids: torch.Tensor, attention_mask_minibatch: torch.Tensor) -> torch.Tensor:
        """
        Forward method of the fine-tuning model.

        Parameters
        ----------
        input_ids : torch.Tensor
            The input ids to the fine-tuning model.
        attention_mask_minibatch : torch.Tensor
            The attention mask for the input tensor.
        
        Returns
        -------
        torch.Tensor
            The output tensor of the fine-tuning model.
        """
        outputs = self.model(input_ids=input_ids, attention_mask=attention_mask_minibatch)
        final_layer = outputs.hidden_states[-1]
        cls_seq = final_layer[:, 0, :]
        final = self.fine_tuning_head(cls_seq)
        return final
    
    def train(
            self,
            train_dataset: Dataset, 
            optimizer: optim = optim.AdamW,
            optimizer_params: dict = {'lr': 0.0001}, 
            loss_function: loss = loss.CrossEntropyLoss(), 
            label: str = "cell_types", 
            epochs: int = 1,
            freeze_layers: int = 2,
            validation_dataset: Optional[Dataset] = None,
            lr_scheduler_params: Optional[dict] = None,
            silent = False):
        """
        Fine-tunes the Geneformer model for classification tasks.

        Parameters
        ----------
        train_dataset : Dataset
            A helical-processed dataset for fine-tuning.

        optimizer : torch.optim.Optimizer, optional, default=torch.optim.AdamW
            The optimizer to be used for training.

        optimizer_params : dict
            Parameters to be passed to the specified optimizer. This dictionary should *NOT* include model parameters.
            Example:
                optimizer_params = {'lr': 0.0001}

        loss_function : torch.nn.loss, optional, default=torch.nn.CrossEntropyLoss()
            The loss function to be used for training.

        label : str, optional, default="cell_types"
            The column in the dataset containing the training labels. Labels should be stored as unique integers per class.

        epochs : int, optional, default=10
            The number of epochs to train the model.

        freeze_layers : int, optional, default=2
            The number of layers to freeze during training.

        validation_dataset : Dataset, optional, default=None
            A helical-processed dataset used for per-epoch validation. If not specified, no validation will be performed.

        lr_scheduler_params : dict, optional, default=None
            Parameters for the learning rate scheduler from the Transformers [`get_scheduler`](https://huggingface.co/docs/transformers/en/main_classes/optimizer_schedules#transformers.get_scheduler) method. The optimizer should not be included 
            in this dictionary, as it will be inferred from the [`optimizer`](https://pytorch.org/docs/main/optim.html) parameter. If not specified, no learning rate scheduler 
            will be used.
            Example:
                lr_scheduler_params = {'name': 'linear', 'num_warmup_steps': 0, 'num_training_steps': 5}
        """

                
        model_input_size = get_model_input_size(self.model)

        cls_present = any("<cls>" in key for key in self.gene_token_dict.keys())
        eos_present = any("<eos>" in key for key in self.gene_token_dict.keys())
        if self.emb_mode == "cls":
            if cls_present is False:
                message = "<cls> token missing in token dictionary"
                logger.error(message)
                raise ValueError(message)
            # Check to make sure that the first token of the filtered input data is cls token
            cls_token_id = self.gene_token_dict["<cls>"]
            if cls_token_id != train_dataset["input_ids"][0][0]:
                message = "First token is not <cls> token value"
                logger.error(message)
            assert (
                train_dataset["input_ids"][0][0] == cls_token_id
            ), "First token is not <cls> token value"
        elif self.emb_mode == "cell":
            if cls_present:
                logger.warning(
                    "CLS token present in token dictionary, excluding from average."
                )
            if eos_present:
                logger.warning(
                    "EOS token present in token dictionary, excluding from average."
                )
        
        total_batch_length = len(train_dataset)
        #initialise optimizer
        optimizer = optimizer(self.parameters(), **optimizer_params)

        #initialise lr_scheduler
        lr_scheduler = None
        if lr_scheduler_params is not None: 
            lr_scheduler = get_scheduler(optimizer=optimizer, **lr_scheduler_params)
        
        if freeze_layers > 0:
            logger.info(f"Freezing the first {freeze_layers} encoder layers of the Geneformer model during fine-tuning.")

            frozen_layers = self.model.bert.encoder.layer[:freeze_layers]

            for module in frozen_layers:
                for param in module.parameters():
                    param.requires_grad = False

        self.to(self.device)

        validation_batch_length = 0
        if validation_dataset is not None:
            validation_batch_length = len(validation_dataset)

        logger.info("Starting Fine-Tuning")
        for j in range(epochs):
            training_loop = trange(0, total_batch_length, self.config["batch_size"], desc="Fine-Tuning", leave=(not silent))
            batch_loss = 0.0
            batches_processed = 0
            for i in training_loop:
                max_range = min(i + self.config["batch_size"], total_batch_length)

                minibatch = train_dataset.select([i for i in range(i, max_range)])
                max_len = int(max(minibatch["length"]))
                minibatch.set_format(type="torch",device=self.device)

                input_data_minibatch = minibatch["input_ids"]
                input_data_minibatch = pad_tensor_list(
                    input_data_minibatch, max_len, self.pad_token_id, model_input_size
                ).to(self.device)

                outputs = self._forward(input_ids=input_data_minibatch, attention_mask_minibatch=gen_attention_mask(minibatch))
                loss = loss_function(outputs, minibatch[label])
                loss.backward()
                batch_loss += loss.item()
                batches_processed += 1
                training_loop.set_postfix({"loss": batch_loss/batches_processed})
                training_loop.set_description(f"Fine-Tuning: epoch {j+1}/{epochs}")
                optimizer.step()
                optimizer.zero_grad()

                del outputs
                del minibatch
                del input_data_minibatch
            if lr_scheduler is not None:
                lr_scheduler.step()

            if validation_dataset is not None:
                testing_loop = trange(0, validation_batch_length, self.config["batch_size"], desc="Fine-Tuning Validation", leave=(not silent))
                val_loss = 0.0
                count = 0.0
                for i in testing_loop:
                    max_range = min(i + self.config["batch_size"], validation_batch_length)

                    minibatch = validation_dataset.select([i for i in range(i, max_range)])
                    max_len = int(max(minibatch["length"]))
                    minibatch.set_format(type="torch",device=self.device)

                    input_data_minibatch = minibatch["input_ids"]
                    input_data_minibatch = pad_tensor_list(
                        input_data_minibatch, max_len, self.pad_token_id, model_input_size
                    ).to(self.device)

                    with torch.no_grad():
                        outputs = self._forward(input_ids=input_data_minibatch, attention_mask_minibatch=gen_attention_mask(minibatch))
                    val_loss += loss_function(outputs, minibatch[label]).item()
                    count += 1.0
                    testing_loop.set_postfix({"val_loss": val_loss/count})

                    del outputs
                    del minibatch
                    del input_data_minibatch
        logger.info(f"Fine-Tuning Complete. Epochs: {epochs}")

    def get_outputs(
        self,
        dataset: Dataset,
        silent = False
    ):
        """Predicts the labels for a dataset using the fine-tuned model.

        Parameters
        ----------
        dataset : Dataset
            The processed dataset to generate outputs for.

        Returns
        -------
        np.ndarray
            The predicted labels in the form of a numpy array
        """
        model_input_size = get_model_input_size(self.model)
        self.to(self.device)

        dataset_length = len(dataset)

        cls_present = any("<cls>" in key for key in self.gene_token_dict.keys())
        eos_present = any("<eos>" in key for key in self.gene_token_dict.keys())
        if self.emb_mode == "cls":
            if cls_present is False:
                message = "<cls> token missing in token dictionary"
                logger.error(message)
                raise ValueError(message)
            assert cls_present, "<cls> token missing in token dictionary"
            # Check to make sure that the first token of the filtered input data is cls token
            cls_token_id = self.gene_token_dict["<cls>"]
            if cls_token_id != dataset["input_ids"][0][0]:
                message = "First token is not <cls> token value"
                logger.error(message)
            assert (
                dataset["input_ids"][0][0] == cls_token_id
            ), "First token is not <cls> token value"
        elif self.emb_mode == "cell":
            if cls_present:
                logger.warning(
                    "CLS token present in token dictionary, excluding from average."
                )
            if eos_present:
                logger.warning(
                    "EOS token present in token dictionary, excluding from average."
                )
        
        output = []
        testing_loop = trange(0, dataset_length, self.config["batch_size"], desc="Generating Outputs", leave=(not silent))
        for i in testing_loop:
            max_range = min(i + self.config["batch_size"], dataset_length)

            minibatch = dataset.select([i for i in range(i, max_range)])
            max_len = int(max(minibatch["length"]))
            minibatch.set_format(type="torch",device=self.device)

            input_data_minibatch = minibatch["input_ids"]
            input_data_minibatch = pad_tensor_list(
                input_data_minibatch, max_len, self.pad_token_id, model_input_size
            ).to(self.device)

            with torch.no_grad():
                outputs = self._forward(input_ids=input_data_minibatch, attention_mask_minibatch=gen_attention_mask(minibatch))
                output.append(outputs.clone().detach())
            del outputs
            del minibatch
            del input_data_minibatch

        return torch.cat(output, dim=0).cpu().numpy()


from setuptools import setup, find_packages

setup(
    name='helical',
    url='https://github.com/helicalAI/helical-package.git',
    author='Benoit Putzeys, Maxime Allard',
    author_email='benoit@helical-ai.com, maxime@helical-ai.com',
    packages=find_packages(),
    install_requires=[
        'requests==2.32.2',
        'pandas==2.2.2',
        'anndata==0.10.7',
        'numpy==1.26.4',
        'scikit-learn>=1.2.2',
        'scipy==1.13.1',
        'gitpython==3.1.43',
        'torch>=2.0.0,<=2.3.0',
        'torchvision>=0.15.0,<=0.18.0',
        'accelerate==0.29.3',
        'transformers==4.45.1',
        'loompy==3.0.7',
        'scib==1.1.5',
        'scikit-misc==0.3.1',
        'torchtext>=0.15.0,<=0.18.0',
        'azure-identity==1.16.0',
        'azure-storage-blob==12.19.1',
        'azure-core==1.30.1',
        'einops==0.8.0',
        'omegaconf==2.3.0',
        'hydra-core==1.3.2',
        'tensorflow>=2.15.0,<=2.17.0',
        'louvain==0.8.2',
        'pyensembl',
        'datasets==2.20.0',
        'mamba-ssm==2.2.2',
        'causal-conv1d==1.4.0'
    ],  
)



from helical.models.base_models import BaseTaskModel
from numpy import ndarray
from sklearn import svm
from typing_extensions import Self
from typing import Optional
import pickle
import os

class SupportVectorMachine(BaseTaskModel):
    def __init__(self, kernel='rbf', degree=3, C=1, decision_function_shape='ovr') -> None:
        self.kernel = kernel
        self.degree = degree
        self.C = C
        self.decision_function_shape = decision_function_shape

    def compile(self, num_classes: Optional[int] = None, input_shape: Optional[int] = None) -> None:
        """
        Compile a SVM. The input parameters are not needed for the SVM model, providing them makes for cleaner code.

        Parameters
        ----------
        num_classes : int, None
            The number of classes to predict, default is None. 
            The SVM should find this automatically.
        input_shape : int, None
            The input shape of the neural network, default is None.
            The SVM should find this automatically.
        """
        self.svm_model = svm.SVC(kernel = self.kernel, 
                                 degree = self.degree, 
                                 C = self.C, 
                                 decision_function_shape = self.decision_function_shape)  

    def train(self, X_train: ndarray, y_train: ndarray, **kwargs) -> Self:
        """Train an SVM on the training.

        Parameters
        ----------
        X_train : ndarray
            The training data features.
        y_train : ndarray
            The training data labels.

        Returns
        -------
        The neural network instance.
        """
        self.svm_model.fit(X_train, y_train)
        return self
  
    def predict(self, x: ndarray) -> ndarray:
        """Use the SVM to make predictions.

        Parameters
        ----------
        x : ndarray
            The data to make predictions upon.

        Returns
        -------
        The prediction of the SVM.
        """        
        return self.svm_model.predict(x)
    
    def save(self, path: str) -> None:
        """Save the SVM model to a file.

        Parameters
        ----------
        path : str
            The path to save the model.
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)  
        file = f"{path}svm.h5"
        with open(file, 'wb') as f:
            pickle.dump(self.svm_model, f)

    def load(self, path: str) -> Self:
        """Load the SVM model from a file.

        Parameters
        ----------
        path : str
            The path to load the model from.

        Returns
        -------
        The SVM instance.
        """
        with open(path, 'rb') as f:
            self.svm_model = pickle.load(f)
        return self
    


import torch

from .model_dir import TransformerModel
from .tokenizer import GeneVocab
from .utils import load_pretrained
from helical.models.scgpt.scgpt_config import scGPTConfig

def load_model(model_configs: scGPTConfig):

    # LOAD MODEL
    model_dir = model_configs["model_path"].parent
    vocab_file = model_dir / "vocab.json"
    special_tokens = [model_configs["pad_token"], "<cls>", "<eoc>"]

    # vocabulary
    vocab = GeneVocab.from_file(vocab_file)
    for s in special_tokens:
        if s not in vocab:
            vocab.append_token(s)

    # Binning will be applied after tokenization. A possible way to do is to use the unified way of binning in the data collator.

    vocab.set_default_index(vocab[model_configs["pad_token"]])

    model = TransformerModel(
        ntoken=len(vocab),
        d_model=model_configs["embsize"],
        nhead=model_configs["nheads"],
        d_hid=model_configs["d_hid"],
        nlayers=model_configs["nlayers"],
        nlayers_cls=model_configs["n_layers_cls"],
        n_cls=1,
        vocab=vocab,
        dropout=model_configs["dropout"],
        pad_token=model_configs["pad_token"],
        pad_value=model_configs["pad_value"],
        do_mvc=True,
        do_dab=False,
        use_batch_labels=False,
        domain_spec_batchnorm=False,
        explicit_zero_prob=False,
        use_fast_transformer=model_configs["use_fast_transformer"],
        fast_transformer_backend="flash",
        pre_norm=False,
    )

    load_pretrained(model, torch.load(model_configs["model_path"], map_location = model_configs["device"]), verbose = False)
    model.to(model_configs["device"])
    model.eval()
    return model, vocab
